# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import os
from typing import Dict, List

HYPERPOD_RECIPE_PATH = os.path.join(
    "sagemaker_hyperpod_recipes",
    "recipes_collection",
    "recipes",
)


class EvaluationTask(enum.Enum):
    """Enum for evaluation tasks."""

    MMLU = "mmlu"
    MMLU_PRO = "mmlu_pro"
    BBH = "bbh"
    GPQA = "gpqa"
    MATH = "math"
    STRONG_REJECT = "strong_reject"
    GEN_QA = "gen_qa"
    IFEVAL = "ifeval"
    MMMU = "mmmu"
    LLM_JUDGE = "llm_judge"
    RUBRIC_LLM_JUDGE = "rubric_llm_judge"
    RFT_EVAL = "rft_eval"
    RFT_MULTITURN_EVAL = "rft_multiturn_eval"  # Maps to "rft_eval" in recipes

    def get_recipe_value(self) -> str:
        """Get the value to use in recipe templates."""
        if self == EvaluationTask.RFT_MULTITURN_EVAL:
            return "rft_eval"
        return self.value


class EvaluationStrategy(enum.Enum):
    """Enum for evaluation strategies."""

    ZERO_SHOT = "zs"
    ZERO_SHOT_COT = "zs_cot"
    FEW_SHOT = "fs"
    FEW_SHOT_COT = "fs_cot"
    GEN_QA = "gen_qa"  # Strategy specific for bring your own dataset.
    JUDGE = "judge"  # Strategy specific for Nova LLM as Judge and rubric_llm_judge.
    RFT_EVAL = "rft_eval"  # Strategy specific for RFT eval
    RFT_MULTITURN_EVAL = "rft_eval"  # Strategy specific for RFT multiturn eval


class EvaluationMetric(enum.Enum):
    """Enum for evaluation metrics."""

    ACCURACY = "accuracy"
    EXACT_MATCH = "exact_match"
    DEFLECTION = "deflection"
    ALL = "all"


EVAL_TASK_STRATEGY_MAP: Dict[EvaluationTask, EvaluationStrategy] = {
    EvaluationTask.MMLU: EvaluationStrategy.ZERO_SHOT_COT,
    EvaluationTask.MMLU_PRO: EvaluationStrategy.ZERO_SHOT_COT,
    EvaluationTask.BBH: EvaluationStrategy.FEW_SHOT_COT,
    EvaluationTask.GPQA: EvaluationStrategy.ZERO_SHOT_COT,
    EvaluationTask.MATH: EvaluationStrategy.ZERO_SHOT_COT,
    EvaluationTask.STRONG_REJECT: EvaluationStrategy.ZERO_SHOT,
    EvaluationTask.IFEVAL: EvaluationStrategy.ZERO_SHOT,
    EvaluationTask.GEN_QA: EvaluationStrategy.GEN_QA,
    EvaluationTask.MMMU: EvaluationStrategy.ZERO_SHOT_COT,
    EvaluationTask.LLM_JUDGE: EvaluationStrategy.JUDGE,
    EvaluationTask.RUBRIC_LLM_JUDGE: EvaluationStrategy.JUDGE,
    EvaluationTask.RFT_EVAL: EvaluationStrategy.RFT_EVAL,
    EvaluationTask.RFT_MULTITURN_EVAL: EvaluationStrategy.RFT_MULTITURN_EVAL,
}

EVAL_TASK_METRIC_MAP: Dict[EvaluationTask, EvaluationMetric] = {
    EvaluationTask.MMLU: EvaluationMetric.ACCURACY,
    EvaluationTask.MMLU_PRO: EvaluationMetric.ACCURACY,
    EvaluationTask.BBH: EvaluationMetric.ACCURACY,
    EvaluationTask.GPQA: EvaluationMetric.ACCURACY,
    EvaluationTask.MATH: EvaluationMetric.EXACT_MATCH,
    EvaluationTask.STRONG_REJECT: EvaluationMetric.DEFLECTION,
    EvaluationTask.IFEVAL: EvaluationMetric.ACCURACY,
    EvaluationTask.GEN_QA: EvaluationMetric.ALL,
    EvaluationTask.MMMU: EvaluationMetric.ACCURACY,
    EvaluationTask.LLM_JUDGE: EvaluationMetric.ALL,
    EvaluationTask.RUBRIC_LLM_JUDGE: EvaluationMetric.ALL,
    EvaluationTask.RFT_EVAL: EvaluationMetric.ALL,
    EvaluationTask.RFT_MULTITURN_EVAL: EvaluationMetric.ALL,
}

# Available subtasks for each evaluation task (from nova_evaluator.py)
EVAL_AVAILABLE_SUBTASKS: Dict[EvaluationTask, List[str]] = {
    EvaluationTask.MMLU: [
        "abstract_algebra",
        "anatomy",
        "astronomy",
        "business_ethics",
        "clinical_knowledge",
        "college_biology",
        "college_chemistry",
        "college_computer_science",
        "college_mathematics",
        "college_medicine",
        "college_physics",
        "computer_security",
        "conceptual_physics",
        "econometrics",
        "electrical_engineering",
        "elementary_mathematics",
        "formal_logic",
        "global_facts",
        "high_school_biology",
        "high_school_chemistry",
        "high_school_computer_science",
        "high_school_european_history",
        "high_school_geography",
        "high_school_government_and_politics",
        "high_school_macroeconomics",
        "high_school_mathematics",
        "high_school_microeconomics",
        "high_school_physics",
        "high_school_psychology",
        "high_school_statistics",
        "high_school_us_history",
        "high_school_world_history",
        "human_aging",
        "human_sexuality",
        "international_law",
        "jurisprudence",
        "logical_fallacies",
        "machine_learning",
        "management",
        "marketing",
        "medical_genetics",
        "miscellaneous",
        "moral_disputes",
        "moral_scenarios",
        "nutrition",
        "philosophy",
        "prehistory",
        "professional_accounting",
        "professional_law",
        "professional_medicine",
        "professional_psychology",
        "public_relations",
        "security_studies",
        "sociology",
        "us_foreign_policy",
        "virology",
        "world_religions",
    ],
    EvaluationTask.BBH: [
        "boolean_expressions",
        "causal_judgement",
        "date_understanding",
        "disambiguation_qa",
        "dyck_languages",
        "formal_fallacies",
        "geometric_shapes",
        "hyperbaton",
        "logical_deduction_five_objects",
        "logical_deduction_seven_objects",
        "logical_deduction_three_objects",
        "movie_recommendation",
        "multistep_arithmetic_two",
        "navigate",
        "object_counting",
        "penguins_in_a_table",
        "reasoning_about_colored_objects",
        "ruin_names",
        "salient_translation_error_detection",
        "snarks",
        "sports_understanding",
        "temporal_sequences",
        "tracking_shuffled_objects_five_objects",
        "tracking_shuffled_objects_seven_objects",
        "tracking_shuffled_objects_three_objects",
        "web_of_lies",
        "word_sorting",
    ],
    EvaluationTask.MATH: [
        "algebra",
        "counting_and_probability",
        "geometry",
        "intermediate_algebra",
        "number_theory",
        "prealgebra",
        "precalculus",
    ],
    EvaluationTask.STRONG_REJECT: [
        "dangerous_or_sensitive_topics",
        "offensive_language",
    ],
    EvaluationTask.MMMU: [
        "accounting",
        "agriculture",
        "architecture_and_engineering",
        "art",
        "art_theory",
        "basic_medical_science",
        "biology",
        "chemistry",
        "clinical_medicine",
        "computer_science",
        "design",
        "economics",
        "electronics",
        "energy_and_power",
        "finance",
        "geography",
        "history",
        "literature",
        "manage",
        "marketing",
        "materials",
        "math",
        "mechanical_engineering",
        "music",
        "pharmacy",
        "physics",
        "psychology",
        "public_health",
        "sociology",
    ],
}


BYOD_AVAILABLE_EVAL_TASKS: List[str] = [
    EvaluationTask.GEN_QA.value,
    EvaluationTask.LLM_JUDGE.value,
    EvaluationTask.RUBRIC_LLM_JUDGE.value,
    EvaluationTask.RFT_EVAL.value,
    EvaluationTask.RFT_MULTITURN_EVAL.value,
]


def get_available_subtasks(task: EvaluationTask) -> List[str]:
    """
    Get available subtasks for a given evaluation task.

    Args:
        task: The evaluation task

    Returns:
        List of available subtasks for the task
    """
    return EVAL_AVAILABLE_SUBTASKS.get(task, [])



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Main entrypoint for customizing and training Nova models.

This module provides the NovaModelCustomizer class which orchestrates the training process.
"""

import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, cast

import boto3

from amzn_nova_customization_sdk.iam.iam_role_creator import (
    create_bedrock_execution_role,
    create_sagemaker_execution_role,
)
from amzn_nova_customization_sdk.manager.runtime_manager import (
    JobConfig,
    RuntimeManager,
    SMHPRuntimeManager,
    SMTJRuntimeManager,
)
from amzn_nova_customization_sdk.model.model_config import (
    REGION_TO_ESCROW_ACCOUNT_MAPPING,
    DeploymentResult,
    EndpointInfo,
)
from amzn_nova_customization_sdk.model.model_enums import (
    SUPPORTED_DATAMIXING_METHODS,
    DeploymentMode,
    DeployPlatform,
    Model,
    Platform,
    TrainingMethod,
)
from amzn_nova_customization_sdk.model.nova_model_customizer_util import (
    requires_custom_eval_data,
    resolve_model_checkpoint_path,
    set_output_s3_path,
)
from amzn_nova_customization_sdk.model.result import (
    EvaluationResult,
    SMHPEvaluationResult,
    SMHPTrainingResult,
    SMTJBatchInferenceResult,
    SMTJEvaluationResult,
    SMTJTrainingResult,
    TrainingResult,
)
from amzn_nova_customization_sdk.model.result.inference_result import InferenceResult
from amzn_nova_customization_sdk.monitor.log_monitor import CloudWatchLogMonitor
from amzn_nova_customization_sdk.monitor.mlflow_monitor import MLflowMonitor
from amzn_nova_customization_sdk.recipe.recipe_builder import RecipeBuilder
from amzn_nova_customization_sdk.recipe.recipe_config import EvaluationTask
from amzn_nova_customization_sdk.rft_multiturn import RFTMultiturnInfrastructure
from amzn_nova_customization_sdk.util.bedrock import (
    BEDROCK_EXECUTION_ROLE_NAME,
    DEPLOYMENT_ARN_NAME,
    check_existing_deployment,
    delete_existing_deployment,
    get_required_bedrock_deletion_permissions,
    get_required_bedrock_update_permissions,
    invoke_model,
    monitor_model_create,
    update_provisioned_throughput_model,
)
from amzn_nova_customization_sdk.util.data_mixing import DataMixing
from amzn_nova_customization_sdk.util.logging import logger
from amzn_nova_customization_sdk.util.platform_util import (
    detect_platform_from_path,
    validate_platform_compatibility,
)
from amzn_nova_customization_sdk.util.recipe import load_recipe_templates
from amzn_nova_customization_sdk.util.sagemaker import (
    SAGEMAKER_EXECUTION_ROLE_NAME,
    _validate_sagemaker_instance_type_for_model_deployment,
    create_model_and_endpoint_config,
    get_model_artifacts,
    invoke_sagemaker_inference,
    setup_environment_variables,
)
from amzn_nova_customization_sdk.validation.endpoint_validator import (
    SAGEMAKER_ENDPOINT_ARN_REGEX,
    validate_endpoint_arn,
    validate_sagemaker_environment_variables,
    validate_unit_count,
)
from amzn_nova_customization_sdk.validation.validator import (
    Validator,
)


class NovaModelCustomizer:
    # Configs not documented in __init__
    validation_config = None
    generated_recipe_dir = None

    def __init__(
        self,
        model: Model,
        method: TrainingMethod,
        infra: RuntimeManager,
        data_s3_path: Optional[str] = None,
        output_s3_path: Optional[str] = None,
        model_path: Optional[str] = None,
        validation_config: Optional[Dict[str, bool]] = None,
        generated_recipe_dir: Optional[str] = None,
        mlflow_monitor: Optional[MLflowMonitor] = None,
        deployment_mode: DeploymentMode = DeploymentMode.FAIL_IF_EXISTS,
        data_mixing_enabled: bool = False,
        image_uri: Optional[str] = None,
    ):
        """
        Initializes a NovaModelCustomizer instance.

        Args:
            model: The Nova model to be trained (e.g., NOVA_MICRO)
            method: The fine-tuning method (e.g., SFT_LORA, DPO)
            infra: Runtime infrastructure manager (e.g., SMTJRuntimeManager)
            data_s3_path: S3 path to the training dataset
            output_s3_path: Optional S3 path for output artifacts. If not provided, will be auto-generated
            model_path: Optional S3 path for model path
            validation_config: Optional dict to control validation. Keys: 'iam' (bool), 'infra' (bool).
                             Defaults to {'iam': True, 'infra': True}
            generated_recipe_dir: Optional path to save generated recipe YAMLs
            mlflow_monitor: Optional MLflowMonitor instance for experiment tracking
            deployment_mode: Behavior when deploying to existing endpoint name. Options:
                           FAIL_IF_EXISTS (default), UPDATE_IF_EXISTS
            data_mixing: Enable data mixing. Default is False.
            image_uri: Optional custom ECR image URI to override the default training image.
                      Must be in format: <account>.dkr.ecr.<region>.amazonaws.com/<repository>:<tag>

        Raises:
            ValueError: If region is unsupported or model is invalid
        """
        self.job_id: Optional[str] = (
            None  # This will be set after train/eval method invoked
        )
        self.job_started_time: Optional[datetime] = (
            None  # This will be set after train/eval method invoked
        )
        self.cloud_watch_log_monitor: Optional[CloudWatchLogMonitor] = (
            None  # This will be set after get_logs method invoked
        )

        region = boto3.session.Session().region_name or "us-east-1"
        if region not in REGION_TO_ESCROW_ACCOUNT_MAPPING:
            raise ValueError(
                f"Region '{region}' is not supported for Nova training. "
                f"Supported regions are: {list(REGION_TO_ESCROW_ACCOUNT_MAPPING.keys())}"
            )

        self.region = region
        self._model = model
        self._image_uri = image_uri
        self._method = method
        self.infra = infra
        self.data_s3_path = data_s3_path
        self.model_path = model_path
        self.validation_config = validation_config
        self.deployment_mode = deployment_mode
        self._platform = (
            Platform.SMTJ
            if isinstance(self.infra, SMTJRuntimeManager)
            else Platform.SMHP
        )

        self.instance_type = self.infra.instance_type

        self.output_s3_path = set_output_s3_path(
            region=self.region,
            output_s3_path=output_s3_path,
            kms_key_id=self.infra.kms_key_id,
        )

        self.generated_recipe_dir = generated_recipe_dir
        self.mlflow_monitor = mlflow_monitor

        # Initialize data mixing configuration
        self.data_mixing_enabled = data_mixing_enabled
        self.data_mixing = None
        if data_mixing_enabled:
            self.data_mixing = DataMixing()
            self._init_data_mixing(self.model, self.method, self.platform)

        self.endpoint_info: Optional[EndpointInfo] = None

    @property
    def model(self) -> Model:
        """Get the model attribute."""
        return self._model

    @model.setter
    def model(self, value: Model) -> None:
        """Set the model attribute and reinitialize data mixing if enabled."""
        if self.data_mixing_enabled:
            self._init_data_mixing(
                model=value, method=self.method, platform=self.platform
            )
            logger.info(
                f"Model changed to {value.name}. Datamixing configs set to default."
            )
        self._model = value

    @property
    def method(self) -> TrainingMethod:
        """Get the method attribute."""
        return self._method

    @method.setter
    def method(self, value: TrainingMethod) -> None:
        """Set the method attribute and reinitialize data mixing if enabled."""
        if self.data_mixing_enabled:
            self._init_data_mixing(
                model=self.model, method=value, platform=self.platform
            )
            logger.info(
                f"Method changed to {value.name}. Datamixing configs set to default."
            )
        self._method = value

    @property
    def platform(self) -> Platform:
        """Get the platform attribute."""
        return self._platform

    @platform.setter
    def platform(self, value: Platform) -> None:
        """Set the platform attribute and reinitialize data mixing if enabled."""
        if self.data_mixing_enabled:
            self._init_data_mixing(model=self.model, method=self.method, platform=value)
            logger.info(
                f"Platform changed to {value.name}. Datamixing configs set to default."
            )
        self._platform = value

    def _init_data_mixing(
        self, model: Model, method: TrainingMethod, platform: Platform
    ) -> None:
        """
        Initialize data mixing configuration.
        """
        if not self.data_mixing_enabled:
            return

        # Data mixing is only supported on HyperPod for certain training methods
        if platform != Platform.SMHP or method not in SUPPORTED_DATAMIXING_METHODS:
            raise ValueError(
                f"Data mixing is only supported for {SUPPORTED_DATAMIXING_METHODS} training methods on SageMaker HyperPod. "
                "Change platform to SMHP or change to a supported training method to use data mixing."
            )

        # Load recipe metadata and templates for non-evaluation methods
        # Eval requires "type" to be passed to load recipes, therefore we load them in evaluate()
        (
            self.recipe_metadata,
            self.recipe_template,
            self.overrides_template,
            self.image_uri,
        ) = load_recipe_templates(
            model=model,
            method=method,
            platform=platform,
            region=self.region,
            data_mixing_enabled=self.data_mixing_enabled,
            instance_type=self.instance_type,
            eval_task=getattr(self, "eval_task", None),
            image_uri_override=self._image_uri,
        )

        # Load default configuration into DataMixing instance if enabled
        if self.data_mixing and self.overrides_template:
            self.data_mixing._load_defaults_from_template(self.overrides_template)

    def get_data_mixing_config(self) -> Dict[str, Any]:
        """
        Get the current data mixing configuration.

        Returns:
            Dictionary containing the data mixing configuration
        """
        if not self.data_mixing:
            return {}
        return self.data_mixing.get_config()

    def set_data_mixing_config(self, config: Dict[str, Any]) -> None:
        """
        Set the data mixing configuration.

        Args:
            config: Dictionary containing the data mixing configuration.
                   Keys should include nova_*_percent fields and customer_data_percent.
                   Any nova_*_percent fields not specified will be set to 0.

        Raises:
            ValueError: If data mixing is not enabled or invalid configuration
        """
        if not self.data_mixing:
            raise ValueError(
                "Data mixing is not enabled for this customizer. Set data_mixing = True in 'NovaModelCustomizer' object."
            )

        self.data_mixing.set_config(config, normalize=True)

    def train(
        self,
        job_name: str,
        recipe_path: Optional[str] = None,
        overrides: Optional[Dict[str, Any]] = None,
        rft_lambda_arn: Optional[str] = None,
        rft_multiturn_infra: Optional[RFTMultiturnInfrastructure] = None,
        validation_data_s3_path: Optional[str] = None,
        dry_run: Optional[bool] = False,
    ) -> TrainingResult | None:
        """
        Generates the recipe YAML, configures runtime, and launches a training job.

        Args:
            job_name: User-defined name for the training job
            recipe_path: Optional path for a YAML recipe file (both S3 and local paths are accepted)
            overrides: Optional dictionary of configuration overrides. Example:
                {
                    'max_epochs': 10,
                    'lr': 5e-6,
                    'warmup_steps': 20,
                    'loraplus_lr_ratio': 16.0,
                    'global_batch_size': 128,
                    'max_length': 16384
                }
            rft_lambda_arn: Optional rewards Lambda ARN, only required for RFT training methods
            rft_multiturn_infra: Optional RFT multiturn infrastructure, required for RFT_MULTITURN methods
            validation_data_s3_path: Optional validation S3 path, only applicable for CPT (but is still optional for CPT)
            dry_run: Actually starts a job if False, otherwise just performs validation. Default is False.

        Returns:
            TrainingResult: Metadata object containing job ID, method, start time, and model artifacts
            or None if dry_run is enabled

        Raises:
            Exception: If job execution fails
        """
        # Create RecipeBuilder and let it handle all data mixing logic
        recipe_builder = RecipeBuilder(
            region=self.region,
            job_name=job_name,
            platform=self.platform,
            model=self.model,
            method=self.method,
            instance_type=self.infra.instance_type,
            instance_count=self.infra.instance_count,
            infra=self.infra,
            data_s3_path=self.data_s3_path,
            output_s3_path=self.output_s3_path,
            model_path=self.model_path,
            rft_lambda_arn=rft_lambda_arn,
            rft_multiturn_infra=rft_multiturn_infra,
            validation_data_s3_path=validation_data_s3_path,
            mlflow_monitor=self.mlflow_monitor,
            data_mixing_instance=self.data_mixing,
            image_uri_override=self._image_uri,
        )

        (
            resolved_recipe_path,
            resolved_output_s3_path,
            resolved_data_s3_path,
            resolved_image_uri,
        ) = recipe_builder.build_and_validate(
            overrides=overrides,
            input_recipe_path=recipe_path,
            output_recipe_path=self.generated_recipe_dir,
            validation_config=self.validation_config,
        )

        if dry_run:
            return None

        # Use unique name to actually start the job
        unique_job_name = f"{job_name}-{uuid.uuid4()}"[:63]

        start_time = datetime.now(timezone.utc)
        self.job_started_time = start_time

        self.job_id = self.infra.execute(
            job_config=JobConfig(
                job_name=unique_job_name,
                data_s3_path=resolved_data_s3_path,
                output_s3_path=resolved_output_s3_path,
                image_uri=resolved_image_uri,
                recipe_path=resolved_recipe_path,
                input_s3_data_type="Converse"
                if self.method not in (TrainingMethod.RFT_LORA, TrainingMethod.RFT_FULL)
                else None,
            )
        )

        training_result: TrainingResult
        if self.platform is Platform.SMTJ:
            training_result = SMTJTrainingResult(
                job_id=self.job_id,
                started_time=start_time,
                method=self.method,
                model_type=self.model,
                model_artifacts=get_model_artifacts(
                    job_name=unique_job_name,
                    infra=self.infra,
                    output_s3_path=resolved_output_s3_path,
                ),
            )
        else:
            cluster_name = cast(SMHPRuntimeManager, self.infra).cluster_name
            namespace = cast(SMHPRuntimeManager, self.infra).namespace
            training_result = SMHPTrainingResult(
                job_id=self.job_id,
                started_time=start_time,
                method=self.method,
                model_type=self.model,
                model_artifacts=get_model_artifacts(
                    job_name=unique_job_name,
                    infra=self.infra,
                    output_s3_path=resolved_output_s3_path,
                ),
                cluster_name=cluster_name,
                namespace=namespace,
            )

        logger.info(f"Started job '{training_result.job_id}'.")
        if training_result.model_artifacts.checkpoint_s3_path:
            logger.info(
                f"Checkpoint S3 path is: {training_result.model_artifacts.checkpoint_s3_path}."
            )
        if training_result.model_artifacts.output_s3_path:
            logger.info(
                f"Output S3 path is: {training_result.model_artifacts.output_s3_path}."
            )

        return training_result

    def evaluate(
        self,
        job_name: str,
        eval_task: EvaluationTask,
        model_path: Optional[str] = None,
        subtask: Optional[str] = None,
        data_s3_path: Optional[str] = None,
        recipe_path: Optional[str] = None,
        overrides: Optional[Dict[str, Any]] = None,
        processor: Optional[Dict[str, Any]] = None,
        rl_env: Optional[Dict[str, Any]] = None,
        rft_multiturn_infra: Optional["RFTMultiturnInfrastructure"] = None,
        dry_run: Optional[bool] = False,
        job_result: Optional[TrainingResult] = None,
    ) -> EvaluationResult | None:
        """
        Generates the recipe YAML, configures runtime, and launches an evaluation job.

        :param job_name: User-defined name for the evaluation job
        :param eval_task: The evaluation task to be performed, e.g. mmlu
        :param model_path: Optional S3 path for model path
        :param subtask: Optional subtask for evaluation
        :param data_s3_path: Optional S3 URI for the dataset
        :param recipe_path: Optional path for a YAML recipe file (both S3 and local paths are accepted)
        :param overrides: Optional dictionary of configuration overrides for eval job (Inference config). Example:
                {
                    'max_new_tokens': 2048,
                    'top_k': -1,
                    'top_p': 1.0,
                    'temperature': 0,
                    'top_logprobs': 10
                }
        :param processor: Optional, only needed for Bring Your Own Metrics Configuration. Example:
                {
                    'lambda_arn': 'arn:aws:lambda:<region>:<account_id>:function:<function-name>',
                    'preprocessing': { # Optional, default to True if not provided
                        'enabled': True
                    },
                    'postprocessing': { # Optional, default to True if not provided
                        'enabled': True
                    },
                    # Built-in aggregation function (valid options: min, max, average, sum), default to average
                    'aggregation': 'average'
                }
        :param rl_env: Optional, only needed for Bring your own Reinforcement learning environment (RFT Eval) config.
                Example:
                {
                    'reward_lambda_arn': 'arn:aws:lambda:<region>:<account_id>:function:<reward-function-name>'
                }
        :param rft_multiturn_infra: Optional RFT multiturn infrastructure, required for RFT_MULTITURN_EVAL
        :param dry_run: dry_run: Actually starts a job if False, otherwise just performs validation. Default is False.
        :param job_result: Optional TrainingResult object to extract checkpoint path from.
                          If provided and model_path is None, will automatically extract
                          the checkpoint path from the training job's output.
        :return: EvaluationResult: Metadata object containing job ID, start time, and evaluation output path
                 or None if dry_run is enabled
        """

        # Resolve model checkpoint path
        resolved_model_path = resolve_model_checkpoint_path(
            model_path=model_path,
            job_result=job_result,
            customizer_job_id=self.job_id if hasattr(self, "job_id") else None,
            customizer_output_s3_path=self.output_s3_path,
            customizer_model_path=self.model_path,
        )

        if resolved_model_path is None:
            logger.warning(
                f"Could not resolve model checkpoint path for evaluate job! Falling back to base model {self.model}"
            )

        # Validate platform compatibility
        checkpoint_platform = None
        if resolved_model_path and resolved_model_path.startswith("s3://"):
            checkpoint_platform = detect_platform_from_path(resolved_model_path)

        if checkpoint_platform is None:
            if job_result is not None:
                if job_result.model_artifacts.checkpoint_s3_path:
                    checkpoint_platform = detect_platform_from_path(
                        job_result.model_artifacts.checkpoint_s3_path
                    )
            elif self.output_s3_path and self.output_s3_path.startswith("s3://"):
                checkpoint_platform = detect_platform_from_path(self.output_s3_path)

        validate_platform_compatibility(
            checkpoint_platform=checkpoint_platform,
            execution_platform=self.platform,
            checkpoint_source="evaluation model checkpoint",
        )

        # Only use the cached data_s3_path for BYOD eval tasks
        if requires_custom_eval_data(eval_task):
            customizer_data_s3_path = self.data_s3_path
        else:
            logger.info(
                f"{eval_task} does not use custom data, ignoring customizer data_s3_path."
            )
            customizer_data_s3_path = None

        recipe_builder = RecipeBuilder(
            region=self.region,
            job_name=job_name,
            platform=self.platform,
            model=self.model,
            method=self.method,
            instance_type=self.infra.instance_type,
            instance_count=self.infra.instance_count,
            infra=self.infra,
            data_s3_path=data_s3_path or customizer_data_s3_path,
            output_s3_path=self.output_s3_path,
            model_path=resolved_model_path,
            eval_task=eval_task,
            subtask=subtask,
            processor_config=processor,
            rl_env_config=rl_env,
            rft_multiturn_infra=rft_multiturn_infra,
            mlflow_monitor=self.mlflow_monitor,
            image_uri_override=self._image_uri,
        )

        (
            resolved_recipe_path,
            resolved_output_s3_path,
            resolved_data_s3_path,
            resolved_image_uri,
        ) = recipe_builder.build_and_validate(
            overrides=overrides,
            input_recipe_path=recipe_path,
            output_recipe_path=self.generated_recipe_dir,
            validation_config=self.validation_config,
        )

        if dry_run:
            return None

        # Use unique name to actually start the job
        unique_job_name = f"{job_name}-{uuid.uuid4()}"[:63]

        start_time = datetime.now(timezone.utc)
        self.job_started_time = start_time

        self.job_id = self.infra.execute(
            job_config=JobConfig(
                job_name=unique_job_name,
                data_s3_path=resolved_data_s3_path,
                output_s3_path=resolved_output_s3_path,
                image_uri=resolved_image_uri,
                recipe_path=resolved_recipe_path,
                input_s3_data_type="S3Prefix",
            )
        )

        evaluation_result: EvaluationResult
        if self.platform == Platform.SMTJ:
            eval_output_s3_path = f"{resolved_output_s3_path.rstrip('/')}/{self.job_id}/output/output.tar.gz"
            evaluation_result = SMTJEvaluationResult(
                job_id=self.job_id,
                eval_task=eval_task,
                started_time=start_time,
                eval_output_path=eval_output_s3_path,
            )
        else:
            cluster_name = cast(SMHPRuntimeManager, self.infra).cluster_name
            namespace = cast(SMHPRuntimeManager, self.infra).namespace
            eval_output_s3_path = (
                f"{resolved_output_s3_path.rstrip('/')}/{self.job_id}/eval-result/"
            )
            evaluation_result = SMHPEvaluationResult(
                job_id=self.job_id,
                eval_task=eval_task,
                started_time=start_time,
                eval_output_path=eval_output_s3_path,
                cluster_name=cluster_name,
                namespace=namespace,
            )
        logger.info(
            f"Started eval job '{self.job_id}'. Artifacts will be published to {eval_output_s3_path}"
        )

        return evaluation_result

    def deploy(
        self,
        model_artifact_path: Optional[str] = None,
        deploy_platform: DeployPlatform = DeployPlatform.BEDROCK_OD,
        unit_count: Optional[int] = 1,
        endpoint_name: Optional[str] = None,
        job_result: Optional[TrainingResult] = None,
        execution_role_name: Optional[str] = None,
        sagemaker_instance_type: Optional[str] = "ml.p5.48xlarge",
        sagemaker_environment_variables: Optional[Dict[str, Any]] = None,
    ) -> DeploymentResult:
        """
        Deployment method supporting both Bedrock and SageMaker platforms.

        Args:
            model_artifact_path: S3 path to the trained model checkpoint. If not provided, will attempt to extract from job_result or the `job_id` field of the Customizer.
            deploy_platform: Platform to deploy to (Bedrock On-Demand, Provisioned Throughput, or SageMaker)
            unit_count: Used in Bedrock Provisioned Throughput number of PT to purchase or SageMaker number of initial instances
            endpoint_name: Name of the deployed model's endpoint (auto-generated if not provided)
            job_result: Training job result object to use for extracting checkpoint path and validating job completion. Also used to retrieve job_id if it's not provided.
            execution_role_name:  Optional IAM execution role name for Bedrock or SageMaker, defaults to BedrockDeployModelExecutionRole or SageMakerExecutionRoleName. If this role does not exist, it will be created.
            sagemaker_instance_type: Optional EC2 instance type for SageMaker deployment, defaults to ml.p5.48xlarge
            sagemaker_environment_variables: Optional environment variables for model configuration

        Returns:
            DeploymentResult with endpoint information
        """

        validate_unit_count(unit_count)

        resolved_model_artifact_path = resolve_model_checkpoint_path(
            model_path=model_artifact_path,
            job_result=job_result,
            customizer_job_id=self.job_id if hasattr(self, "job_id") else None,
            customizer_output_s3_path=self.output_s3_path,
            customizer_model_path=self.model_path,
            fail_on_error=True,
        )

        if resolved_model_artifact_path is None:
            raise ValueError(
                "Model artifact path could not be resolved. Provide a valid model_path or job_result"
            )

        if deploy_platform in [DeployPlatform.BEDROCK_OD, DeployPlatform.BEDROCK_PT]:
            if execution_role_name is None:
                execution_role_name = BEDROCK_EXECUTION_ROLE_NAME

            return self._deploy_to_bedrock(
                model_artifact_path=resolved_model_artifact_path,
                deploy_platform=deploy_platform,
                pt_units=unit_count,
                endpoint_name=endpoint_name,
                execution_role_name=execution_role_name,
            )

        elif deploy_platform == DeployPlatform.SAGEMAKER:
            if execution_role_name is None:
                execution_role_name = SAGEMAKER_EXECUTION_ROLE_NAME

            if job_result is not None:
                model = job_result.model_type
            elif self.model:
                model = self.model
            else:
                raise ValueError(
                    "model_type must be provided either through job_result in deploy() or during "
                    "NovaModelCustomizer initialization for SageMaker deployment"
                )

            if sagemaker_instance_type is None:
                raise ValueError(
                    "sagemaker_instance_type cannot be None for SageMaker deployment"
                )

            _validate_sagemaker_instance_type_for_model_deployment(
                sagemaker_instance_type, model
            )

            # Path required to end in /
            resolved_model_artifact_path = (
                resolved_model_artifact_path
                if resolved_model_artifact_path.endswith("/")
                else resolved_model_artifact_path + "/"
            )

            return self._deploy_to_sagemaker(
                model_artifact_path=resolved_model_artifact_path,
                endpoint_name=endpoint_name,
                instance_type=sagemaker_instance_type,
                unit_count=unit_count,  # type: ignore
                environment_variables=sagemaker_environment_variables,
                execution_role_name=execution_role_name,
            )

        else:
            raise ValueError(f"Unsupported deployment platform: {deploy_platform}")

    def _deploy_to_sagemaker(
        self,
        model_artifact_path: str,
        instance_type: str,
        unit_count: int,
        endpoint_name: Optional[str] = None,
        environment_variables: Optional[Dict[str, Any]] = None,
        execution_role_name: str = SAGEMAKER_EXECUTION_ROLE_NAME,
    ) -> DeploymentResult:
        """
        Internal method to deploy model to SageMaker.

        Args:
            model_artifact_path: S3 path to model artifacts
            endpoint_name: Custom endpoint name (optional)
            unit_count: Number of instances
            instance_type: SageMaker instance type
            environment_variables: Model configuration variables
            execution_role_name: execution role name

        Returns:
            DeploymentResult with SageMaker endpoint info
        """

        if environment_variables:
            validate_sagemaker_environment_variables(environment_variables)
            env_vars = environment_variables
        else:
            # Use default environment variables if not provided
            env_vars = setup_environment_variables()

        # Generate endpoint name if not provided
        if endpoint_name is None:
            endpoint_name = f"{self.model.value}-{self.method.value}-sagemaker".replace(
                "_", "-"
            ).lower()

        iam_client = boto3.client("iam")

        # Check if a SageMaker IAM execution role exists, if not, create one.
        try:
            sagemaker_role_arn = create_sagemaker_execution_role(
                iam_client=iam_client, role_name=execution_role_name
            )["Role"]["Arn"]
        except Exception as e:
            raise Exception(
                f"Failed to find or create the SageMaker IAM Execution Role: {str(e)}"
            )

        # Create SageMaker clients
        sagemaker_client = boto3.client("sagemaker")

        # Create model, endpoint config, and endpoint
        model_name = f"{endpoint_name}-model"
        endpoint_config_name = f"{endpoint_name}-config"

        endpoint_arn = create_model_and_endpoint_config(
            region=self.region,
            model_name=model_name,
            model_s3_location=model_artifact_path,
            sagemaker_execution_role_arn=sagemaker_role_arn,
            endpoint_config_name=endpoint_config_name,
            endpoint_name=endpoint_name,
            instance_type=instance_type,
            environment=env_vars,
            sagemaker_client=sagemaker_client,
            initial_instance_count=unit_count,
            deployment_mode=self.deployment_mode,
        )

        # Create deployment result
        create_time = datetime.now(timezone.utc)
        endpoint = EndpointInfo(
            platform=DeployPlatform.SAGEMAKER,
            endpoint_name=endpoint_name,
            uri=endpoint_arn,
            model_artifact_path=model_artifact_path,
        )

        self.endpoint_info = endpoint

        return DeploymentResult(endpoint=endpoint, created_at=create_time)

    def _deploy_to_bedrock(
        self,
        model_artifact_path: Optional[str] = None,
        deploy_platform: DeployPlatform = DeployPlatform.BEDROCK_OD,
        pt_units: Optional[int] = None,
        endpoint_name: Optional[str] = None,
        execution_role_name: str = BEDROCK_EXECUTION_ROLE_NAME,
    ) -> DeploymentResult:
        """
        Creates a custom model and deploys it to Bedrock.

        Deployment behavior when endpoint already exists is controlled by the deployment_mode
        parameter set during NovaModelCustomizer initialization:
        - FAIL_IF_EXISTS: Raise error (default, safest)
        - UPDATE_IF_EXISTS: Try in-place update, fail if not supported

        Args:
            model_artifact_path: The s3 path to the training escrow bucket. If not provided, will attempt to extract
                                 from job_result or the `job_id` field of the Customizer.
            deploy_platform: The platform to deploy the model to for inference (Bedrock On-Demand or Provisioned Throughput).
            pt_units: Only needed when Bedrock Provisioned Throughput is chosen. The # of PT to purchase.
            endpoint_name: The name of the deployed model's endpoint -- will be auto generated if not given.
            execution_role_name: Optional IAM execution role name for Bedrock, defaults to BedrockDeployModelExecutionRole. If this role does not exist, it will be created.

        Returns:
            DeploymentResult: Contains the endpoint information as well as the create time of the deployment.

        Raises:
            Exception: When unable to successfully deploy the model, extract checkpoint path, or handle
                      existing endpoint according to deployment_mode setting.
        """
        bedrock_client = boto3.client("bedrock")
        iam_client = boto3.client("iam")

        # Check if we have a model name (endpoint name) else generate one.
        if endpoint_name is None:
            name_format = f"{self.model}-{self.method}-{self.region}".lower()
            endpoint_name = name_format.replace(".", "-").replace("_", "-")

        # Check for existing deployment with same name
        existing_deployment_arn = check_existing_deployment(
            endpoint_name, deploy_platform
        )
        deleted_existing_deployment = False
        should_delete_existing = False
        attempt_pt_update = False

        # Handle cases where the given endpoint name already has an associated deployment
        if existing_deployment_arn:
            if self.deployment_mode == DeploymentMode.FAIL_IF_EXISTS:
                raise Exception(
                    f"Deployment '{endpoint_name}' already exists on platform {deploy_platform}.\n"
                    f"ARN: {existing_deployment_arn}\n"
                    f"Change deployment_mode to override."
                )

            elif self.deployment_mode == DeploymentMode.UPDATE_IF_EXISTS:
                if deploy_platform != DeployPlatform.BEDROCK_PT:
                    raise Exception(
                        f"UPDATE_IF_EXISTS mode is only supported for Provisioned Throughput deployments.\n"
                        f"Current platform: {deploy_platform}\n"
                        f"Use FORCE_REPLACE mode for On-Demand deployments."
                    )
                logger.info(
                    f"UPDATE_IF_EXISTS mode: Will update existing PT deployment '{endpoint_name}' in-place"
                )
                attempt_pt_update = True

            elif self.deployment_mode in [
                DeploymentMode.UPDATE_OR_REPLACE,
                DeploymentMode.FORCE_REPLACE,
            ]:
                # For PT deployments, try in-place update first (unless FORCE_REPLACE)
                if (
                    deploy_platform == DeployPlatform.BEDROCK_PT
                    and self.deployment_mode == DeploymentMode.UPDATE_OR_REPLACE
                ):
                    logger.info(
                        f"UPDATE_OR_REPLACE mode: Will try to update existing PT deployment '{endpoint_name}' in-place"
                    )
                    attempt_pt_update = True

                # Always mark for deletion as fallback (or primary for FORCE_REPLACE/OD)
                if (
                    not attempt_pt_update
                    or self.deployment_mode == DeploymentMode.FORCE_REPLACE
                ):
                    logger.info(
                        f"{self.deployment_mode.value} mode: Will delete existing deployment '{endpoint_name}'"
                    )
                    should_delete_existing = True

        # Consolidated permission validation (if IAM validation enabled)
        if (
            self.validation_config is None or self.validation_config.get("iam", True)
        ) and existing_deployment_arn:
            if attempt_pt_update:
                required_perms = get_required_bedrock_update_permissions(
                    deploy_platform, existing_deployment_arn
                )
                errors: List[str] = []
                Validator._validate_calling_role_permissions(
                    errors, required_perms, infra=None, region_name=self.region
                )
                if errors:
                    if self.deployment_mode == DeploymentMode.UPDATE_IF_EXISTS:
                        raise Exception(
                            f"Cannot update existing PT deployment '{endpoint_name}': Missing permissions.\n"
                            f"{'; '.join(errors)}\n"
                            f"Please ensure your role has the necessary Bedrock update permissions."
                        )
                    else:
                        # UPDATE_OR_REPLACE: fall back to delete
                        logger.warning(
                            f"Missing update permissions, will fall back to delete/recreate"
                        )
                        attempt_pt_update = False
                        should_delete_existing = True

            if should_delete_existing:
                required_perms = get_required_bedrock_deletion_permissions(
                    deploy_platform, existing_deployment_arn
                )
                errors = []
                Validator._validate_calling_role_permissions(
                    errors, required_perms, infra=None, region_name=self.region
                )
                if errors:
                    raise Exception(
                        f"Cannot delete existing deployment '{endpoint_name}': Missing permissions.\n"
                        f"{'; '.join(errors)}\n"
                        f"Please ensure your role has the necessary Bedrock deletion permissions."
                    )

        # TODO: If given a job ID, check the status before creating the model. If the job isn't completed, tell the user.
        # TODO: If a user already has an arn of a custom model, they should be able to directly deploy it.

        # Check if a Bedrock IAM execution role exists, if not, create one.
        try:
            bedrock_execution_role_arn = create_bedrock_execution_role(
                iam_client=iam_client, role_name=execution_role_name
            )["Role"]["Arn"]
        except Exception as e:
            raise Exception(
                f"Failed to find or create the Bedrock IAM Execution Role: {str(e)}"
            )

        model_name = None

        modelKmsKeyArn = None

        try:
            logger.info(f"Creating custom model for endpoint '{endpoint_name}'...")
            model_name = f"{endpoint_name}-{uuid.uuid4()}"[:63]
            if self.infra.kms_key_id:
                sts_client = boto3.client("sts")
                account_id = sts_client.get_caller_identity()["Account"]
                modelKmsKeyArn = f"arn:aws:kms:{self.region}:{account_id}:key/{self.infra.kms_key_id}"
                model = bedrock_client.create_custom_model(
                    modelName=model_name,
                    modelSourceConfig={"s3DataSource": {"s3Uri": model_artifact_path}},
                    roleArn=bedrock_execution_role_arn,
                    modelKmsKeyArn=modelKmsKeyArn,
                )
            else:
                model = bedrock_client.create_custom_model(
                    modelName=model_name,
                    modelSourceConfig={"s3DataSource": {"s3Uri": model_artifact_path}},
                    roleArn=bedrock_execution_role_arn,
                )
        except Exception as e:
            raise Exception(
                f"Failed to create model {model_name} for endpoint {endpoint_name}: {e}"
            )

        # Monitor the model's creation, updating the time stamp every few seconds until the model is created/set as 'active'.
        try:
            monitor_model_create(bedrock_client, model, endpoint_name)
        except Exception as e:
            raise Exception(
                f"Failed to create deployment {endpoint_name} for model {model['modelArn']}: {e}"
            )

        # Delete existing deployment if needed (after model creation succeeds)
        if existing_deployment_arn and should_delete_existing:
            try:
                delete_existing_deployment(
                    existing_deployment_arn, deploy_platform, endpoint_name
                )
                deleted_existing_deployment = True
            except Exception as e:
                raise Exception(
                    f"Failed to create deployment {endpoint_name} for model {model['modelArn']}: Could not delete existing deployment: {e}"
                )

        # Handle deployment creation or update
        deployment = None
        deployment_arn = None
        pt_update_error = None

        # Try PT update if applicable
        if attempt_pt_update and existing_deployment_arn:
            try:
                update_provisioned_throughput_model(
                    existing_deployment_arn, model["modelArn"], endpoint_name
                )
                deployment_arn = existing_deployment_arn
                logger.info(
                    f"Successfully updated existing PT deployment '{endpoint_name}'"
                )
            except Exception as e:
                pt_update_error = str(e)
                if self.deployment_mode == DeploymentMode.UPDATE_IF_EXISTS:
                    raise Exception(
                        f"Failed to create deployment {endpoint_name} for model {model['modelArn']}: {e}"
                    )
                else:
                    # UPDATE_OR_REPLACE: fall back to delete/recreate
                    logger.warning(
                        f"PT update failed, falling back to delete/recreate: {e}"
                    )
                    should_delete_existing = True
                    attempt_pt_update = False

        # Create new deployment if pt update failed or wasn't attempted
        if deployment_arn is None:
            # Delete existing deployment if needed and not already done
            if (
                existing_deployment_arn
                and should_delete_existing
                and not deleted_existing_deployment
            ):
                try:
                    delete_existing_deployment(
                        existing_deployment_arn, deploy_platform, endpoint_name
                    )
                    deleted_existing_deployment = True
                except Exception as e:
                    error_msg = f"Failed to create deployment {endpoint_name} for model {model['modelArn']}: Could not delete existing deployment: {e}"
                    if pt_update_error:
                        error_msg += f". Previous PT update error: {pt_update_error}"
                    raise Exception(error_msg)

            try:
                logger.info(f"Creating deployment for endpoint '{endpoint_name}'...")
                if deploy_platform == DeployPlatform.BEDROCK_PT:
                    deployment = bedrock_client.create_provisioned_model_throughput(
                        modelUnits=pt_units,
                        provisionedModelName=endpoint_name,
                        modelId=model["modelArn"],
                    )
                    deployment_arn = deployment[
                        DEPLOYMENT_ARN_NAME.get(deploy_platform)
                    ]
                elif deploy_platform == DeployPlatform.BEDROCK_OD:
                    deployment = bedrock_client.create_custom_model_deployment(
                        modelDeploymentName=endpoint_name,
                        modelArn=model["modelArn"],
                    )
                    deployment_arn = deployment[
                        DEPLOYMENT_ARN_NAME.get(deploy_platform)
                    ]
                else:
                    raise ValueError(
                        f"Platform '{deploy_platform}' is not supported for Nova training. "
                        f"Supported platforms are: {list(DeployPlatform)}"
                    )
            except Exception as e:
                raise Exception(
                    f"Failed to create deployment {endpoint_name} for model {model['modelArn']}: {e}"
                )

        # Creates EndpointInfo and DeploymentResult objects.
        create_time = datetime.now(timezone.utc)
        self.job_started_time = create_time
        endpoint = EndpointInfo(
            platform=deploy_platform,
            endpoint_name=endpoint_name,
            uri=deployment_arn,
            model_artifact_path=self.output_s3_path,
        )

        self.endpoint_info = endpoint

        result = DeploymentResult(endpoint=endpoint, created_at=create_time)

        # Log message to the user with information about the deployment.
        logger.info(
            f"\nSuccessfully started deploying {endpoint.endpoint_name}: \n"
            f"- Platform: {endpoint.platform}:\n"
            f"- ARN: {endpoint.uri}\n"
            f"- Created: {result.created_at.strftime('%Y-%m-%d %H:%M:%S')} UTC\n"
            f"- ETA: Deployment should be completed in about 30-45 minutes"
        )
        return result

    def predict(self):
        pass

    def batch_inference(
        self,
        job_name: str,
        input_path: str,
        output_s3_path: str,
        model_path: Optional[str] = None,
        recipe_path: Optional[str] = None,
        overrides: Optional[Dict[str, Any]] = None,
        dry_run: Optional[bool] = False,
        job_result: Optional[TrainingResult] = None,
    ) -> InferenceResult | None:
        """
        Launches a batch inference job on a trained model.

        :param job_name: Name for the batch inference job
        :param input_path: S3 path to input data for inference
        :param output_s3_path: S3 path for inference outputs
        :param model_path: Optional S3 path to the model
        :param recipe_path: Optional path for a YAML recipe file
        :param overrides: Optional configuration overrides for inference
        :param dry_run: Actually starts a job if False, otherwise just performs validation. Default is False.
        :param job_result: Optional TrainingResult object to extract checkpoint path from.
                          If provided and model_path is None, will automatically extract
                          the checkpoint path from the training job's output.
        :return: InferenceResult or None if dry_run is enabled
        """

        # Resolve model checkpoint path
        resolved_model_path = resolve_model_checkpoint_path(
            model_path=model_path,
            job_result=job_result,
            customizer_job_id=self.job_id if hasattr(self, "job_id") else None,
            customizer_output_s3_path=self.output_s3_path,
            customizer_model_path=self.model_path,
        )

        if resolved_model_path is None:
            logger.warning(
                f"Could not resolve model checkpoint path for evaluate job! Falling back to base model {self.model}"
            )

        # Validate platform compatibility
        checkpoint_platform = None
        if resolved_model_path and resolved_model_path.startswith("s3://"):
            checkpoint_platform = detect_platform_from_path(resolved_model_path)

        if checkpoint_platform is None and job_result is not None:
            if job_result.model_artifacts.checkpoint_s3_path:
                checkpoint_platform = detect_platform_from_path(
                    job_result.model_artifacts.checkpoint_s3_path
                )

        if (
            checkpoint_platform is None
            and self.model_path
            and self.model_path.startswith("s3://")
        ):
            checkpoint_platform = detect_platform_from_path(self.model_path)

        validate_platform_compatibility(
            checkpoint_platform=checkpoint_platform,
            execution_platform=self.platform,
            checkpoint_source="batch inference model checkpoint",
        )

        recipe_builder = RecipeBuilder(
            region=self.region,
            job_name=job_name,
            platform=self.platform,
            model=self.model,
            method=self.method,
            instance_type=self.infra.instance_type,
            instance_count=self.infra.instance_count,
            infra=self.infra,
            data_s3_path=input_path,
            output_s3_path=output_s3_path or self.output_s3_path,
            model_path=resolved_model_path,
            mlflow_monitor=self.mlflow_monitor,
            image_uri_override=self._image_uri,
        )

        (
            resolved_recipe_path,
            resolved_output_s3_path,
            resolved_data_s3_path,
            resolved_image_uri,
        ) = recipe_builder.build_and_validate(
            overrides=overrides,
            input_recipe_path=recipe_path,
            output_recipe_path=self.generated_recipe_dir,
            validation_config=self.validation_config,
        )

        if dry_run:
            return None

        # Use unique name to actually start the job
        unique_job_name = f"{job_name}-{uuid.uuid4()}"[:63]

        start_time = datetime.now(timezone.utc)
        self.job_started_time = start_time

        self.job_id = self.infra.execute(
            job_config=JobConfig(
                job_name=unique_job_name,
                data_s3_path=resolved_data_s3_path,
                output_s3_path=resolved_output_s3_path,
                image_uri=resolved_image_uri,
                recipe_path=resolved_recipe_path,
                input_s3_data_type="S3Prefix",
            )
        )

        # TODO: Implement for SMHP jobs. I'm not sure how different the infrastructure is.
        inference_output_s3_path = (
            f"{resolved_output_s3_path.rstrip('/')}/{job_name}/output/output.tar.gz"
        )
        batch_inference_result = SMTJBatchInferenceResult(
            job_id=self.job_id,
            started_time=start_time,
            inference_output_path=inference_output_s3_path,
        )
        logger.info(
            f"Started batch inference job '{self.job_id}'. \nArtifacts will be published to {inference_output_s3_path}.\n"
            f"After opening the tar file, look for {recipe_builder.job_name}/eval_results/inference_output.jsonl."
        )
        return batch_inference_result

    def get_logs(
        self,
        limit: Optional[int] = None,
        start_from_head: bool = False,
        end_time: Optional[int] = None,
    ):
        if self.job_id and self.job_started_time:
            kwargs = {}
            if self.platform == Platform.SMHP:
                kwargs["cluster_name"] = cast(
                    SMHPRuntimeManager, self.infra
                ).cluster_name
                kwargs["namespace"] = cast(SMHPRuntimeManager, self.infra).namespace
            self.cloud_watch_log_monitor = (
                self.cloud_watch_log_monitor
                or CloudWatchLogMonitor(
                    job_id=self.job_id,
                    platform=self.platform,
                    started_time=int(self.job_started_time.timestamp() * 1000),
                    **kwargs,
                )
            )
            self.cloud_watch_log_monitor.show_logs(
                limit=limit, start_from_head=start_from_head, end_time=end_time
            )
        else:
            logger.info(
                "No job_id and job_started_time found for this model, please call .train() or .evaluate() first."
            )

    def invoke_inference(
        self, request_body: Dict[str, Any], endpoint_arn: Optional[str] = None
    ):
        """
        Invokes single inference against an endpoint
        :param request_body: Inference request body
        :param endpoint_arn: Optional endpoint ARN if user does not want to use previously deployed endpoint
        :return: Inference result. String for non-streaming response, generator for streaming response
        """
        model_endpoint_arn = ""
        deployment_platform = DeployPlatform.BEDROCK_PT
        if endpoint_arn is None and self.endpoint_info is None:
            raise ValueError(
                "endpoint_arn must be provided if no endpoint was previously deployed by Customizer"
            )
        elif endpoint_arn is not None:
            validate_endpoint_arn(endpoint_arn=endpoint_arn)

            if SAGEMAKER_ENDPOINT_ARN_REGEX.match(endpoint_arn):
                deployment_platform = DeployPlatform.SAGEMAKER

            model_endpoint_arn = endpoint_arn
        elif self.endpoint_info is not None:
            model_endpoint_arn = self.endpoint_info.uri
            deployment_platform = self.endpoint_info.platform

        if deployment_platform == DeployPlatform.SAGEMAKER:
            runtime_client = boto3.client("sagemaker-runtime", region_name=self.region)
            endpoint_name = model_endpoint_arn.split("/")[-1]
            logger.info(f"endpoint name {endpoint_name}")
            return invoke_sagemaker_inference(
                request_body, endpoint_name, runtime_client
            )
        else:
            runtime_client = boto3.client("bedrock-runtime", region_name=self.region)
            return invoke_model(
                model_id=model_endpoint_arn,
                request_body=request_body,
                bedrock_runtime=runtime_client,
            )

    def monitor_metrics(self):
        pass



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .cpt_dataset_validator import CPTDatasetValidator
from .dataset_validator import BaseDatasetValidator
from .eval_dataset_validator import EvalDatasetValidator
from .rft_dataset_validator import RFTDatasetValidator
from .sft_dataset_validator import SFTDatasetValidator

__all__ = [
    "CPTDatasetValidator",
    "BaseDatasetValidator",
    "EvalDatasetValidator",
    "RFTDatasetValidator",
    "SFTDatasetValidator",
]



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import re
import shutil
import tarfile
import tempfile
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, Generator, Optional
from urllib.parse import urlparse

import boto3

from amzn_nova_customization_sdk.model.result.job_result import (
    BaseJobResult,
    JobStatus,
    JobStatusManager,
    SMTJStatusManager,
)
from amzn_nova_customization_sdk.util.logging import logger


@dataclass
class InferenceResult(BaseJobResult, ABC):
    inference_output_path: str

    def __init__(self, job_id: str, started_time: datetime, inference_output_path: str):
        self.inference_output_path = inference_output_path
        self._cached_results_dir: Optional[str] = None
        super().__init__(job_id, started_time)

    @staticmethod
    def _reformat_inference_results(line) -> Dict:
        """
        This reformats each JSONL line into the following format:
        {   "system_prompt":
            "user_response":
            "inference_response":
            "gold_response":
        }

        Args:
            line: This is the line that we want to reformat into the above JSONL format.

        Returns:
            The reformatted JSONL line in the new format.
        """
        entry = json.loads(line.strip())

        # Extract system and user using regex because of more complex structure.
        system_match = re.search(
            r"'role':\s*'system',\s*'content':\s*'([^']+)'", entry["prompt"]
        )
        system_prompt = system_match.group(1) if system_match else ""
        user_match = re.search(
            r"'role':\s*'user',\s*'content':\s*['\"]([^'\"]+)['\"]", entry["prompt"]
        )
        user_response = user_match.group(1) if user_match else ""

        # Clean the inference line
        inference_response = (
            entry["inference"]
            .replace("\\n", "")
            .replace("\\", "")
            .replace("['", "")
            .replace("']", "")
        )
        gold_response = entry["gold"]

        try:
            metadata = entry["metadata"]
        except KeyError:
            metadata = {}

        transformed_entry = {
            "system": system_prompt,
            "query": user_response,
            "gold_response": gold_response,
            "inference_response": inference_response,
        }

        # If metadata is included, append it to the entry before it's returned.
        if metadata:
            transformed_entry.update({"metadata": metadata})

        return transformed_entry

    def _download_inference_results(self) -> str:
        # Check if results are already cached
        if self._cached_results_dir and Path(self._cached_results_dir).exists():
            return self._cached_results_dir
        try:
            # Download and extract results
            logger.info(f"Downloading raw results from {self.inference_output_path}")
            parsed = urlparse(self.inference_output_path)
            bucket = parsed.netloc
            key = parsed.path.lstrip("/")

            s3_client = boto3.client("s3")

            # Create temp dir for caching results
            self._cached_results_dir = tempfile.mkdtemp(
                prefix=f"inference_results_{self.job_id}_"
            )

            local_file = Path(self._cached_results_dir) / "output.tar.gz"
            s3_client.download_file(bucket, key, str(local_file))
            logger.info(
                f"Successfully downloaded inference results to {self._cached_results_dir}"
            )

            # Only extract the inference_output jsonl file.
            with tarfile.open(local_file, "r:gz") as tar:
                for member in tar.getmembers():
                    if member.name.endswith("inference_output.jsonl"):
                        tar.extract(member, self._cached_results_dir)
                        break
            return self._cached_results_dir
        except Exception as e:
            logger.error(
                f"Error retrieving inference results: {e}\n"
                f"Results available at: {self.inference_output_path} in the eval_results folder."
            )
            raise

    # 'print' statements are used here to make them visible in the console when a user calls the "show()" method.
    def get(self, s3_path=None) -> Dict:
        job_status, raw_status = self.get_job_status()
        inference_dataset: list[dict] = []
        if job_status == JobStatus.COMPLETED:
            print(f"Job '{self.job_id}' completed successfully.")
            logger.info(f"Job '{self.job_id}' completed successfully.")

            results_dir = self._download_inference_results()
            results_file = list(Path(results_dir).rglob("inference_output.jsonl"))
            if results_file:
                with open(results_file[0], "r") as f:
                    for line in f:
                        formatted_line = self._reformat_inference_results(line)
                        inference_dataset.append(formatted_line)

                    # If a file path is given, try to save the file to that location.
                    if s3_path:
                        try:
                            jsonl_content = "\n".join(
                                json.dumps(inference, ensure_ascii=False)
                                for inference in inference_dataset
                            )

                            if s3_path.startswith("s3://"):
                                s3_path_stripped = s3_path[5:]  # Remove 's3://'
                                bucket, key = s3_path_stripped.split("/", 1)

                                s3_client = boto3.client("s3")
                                s3_client.put_object(
                                    Bucket=bucket,
                                    Key=key,
                                    Body=jsonl_content.encode("utf-8"),
                                    ContentType="application/jsonlines",
                                )
                            else:
                                local_path = Path(s3_path)
                                local_path.parent.mkdir(parents=True, exist_ok=True)
                                local_path.write_text(jsonl_content, encoding="utf-8")

                            print(f"Successfully saved the results to {s3_path}.")
                            logger.info(f"Successfully saved the results to {s3_path}.")
                        except Exception as e:
                            print(f"Error saving inference results to {s3_path}: {e}")
                            logger.error(
                                f"Error saving inference results to {s3_path}: {e}"
                            )
                            raise
                    return {"inference_results": inference_dataset}
            else:
                print(f"No inference output jsonl file found for job {self.job_id}")
                logger.info(
                    f"No inference output jsonl file found for job {self.job_id}"
                )
                return {}
        elif job_status == JobStatus.IN_PROGRESS:
            print(
                f"Job '{self.job_id}' still running in progress. Please wait until the job is completed."
            )
            logger.info(
                f"Job '{self.job_id}' still running in progress. Please wait until the job is completed."
            )
            return {}
        else:
            print(
                f"Cannot show inference result. Job '{self.job_id}' in {raw_status} status."
            )
            logger.info(
                f"Cannot show inference result. Job '{self.job_id}' in {raw_status} status."
            )
            return {}

    def show(self):
        results = self.get()
        if results:
            print(f"\nInference Results for job_id={self.job_id}:")
            for result in results["inference_results"]:
                print(json.dumps(result, ensure_ascii=False))

    def clean(self):
        if (
            hasattr(self, "_cached_results_dir")
            and self._cached_results_dir
            and Path(self._cached_results_dir).exists()
        ):
            try:
                shutil.rmtree(self._cached_results_dir)
                logger.debug(
                    f"Successfully cleaned temp file in {self._cached_results_dir}"
                )
                self._cached_results_dir = None
            except Exception as e:
                logger.error(
                    f"Failed to remove cached results directory: {self._cached_results_dir} due to {str(e)}"
                )

    def __del__(self):
        self.clean()


@dataclass
class SMTJBatchInferenceResult(InferenceResult):
    def __init__(
        self,
        job_id: str,
        started_time: datetime,
        inference_output_path: str,
        sagemaker_client=None,
    ):
        self._cached_results_dir: Optional[str] = None
        self._sagemaker_client = sagemaker_client or boto3.client("sagemaker")
        super().__init__(job_id, started_time, inference_output_path)

    def _create_status_manager(self) -> JobStatusManager:
        return SMTJStatusManager(self._sagemaker_client)

    def _to_dict(self):
        return {
            "job_id": self.job_id,
            "started_time": self.started_time.isoformat(),
            "inference_output_path": self.inference_output_path,
        }

    @classmethod
    def _from_dict(cls, data) -> "SMTJBatchInferenceResult":
        return cls(
            job_id=data["job_id"],
            started_time=datetime.fromisoformat(data["started_time"]),
            inference_output_path=data["inference_output_path"],
        )


# TODO: Support Multi-modal
@dataclass
class SingleInferenceResult(InferenceResult):
    def __init__(
        self,
        job_id: str,
        started_time: datetime,
        inference_output_path: str,
        nonstreaming_response: Optional[str],
        streaming_response: Optional[Generator[str, None, None]],
    ):
        self._cached_results_dir: Optional[str] = None
        self._nonstreaming_response = nonstreaming_response
        self._streaming_response = streaming_response
        self._is_consumed = False
        super().__init__(job_id, started_time, inference_output_path)

    def _to_dict(self):
        return {
            "job_id": self.job_id,
            "started_time": self.started_time.isoformat(),
            "inference_output_path": self.inference_output_path,
            "nonstreaming_response": self._nonstreaming_response,
            "streaming_response": self._streaming_response,
        }

    def _create_status_manager(self):
        pass

    def get(self, s3_path=None) -> Dict:
        return {
            "inference_results": {
                "is_streaming": True if self._streaming_response is not None else False,
                "response": self._streaming_response
                if self._streaming_response is not None
                else self._nonstreaming_response,
            }
        }

    def show(self):
        """
        Display the streaming results in real-time.

        Warning:
        - This method consumes the generator
        - Subsequent calls to get() will return an empty result
        """

        if self._is_consumed:
            raise ValueError("Streaming response generator has already been consumed")

        results = self.get()
        if results:
            logger.info(f"Inference Results for job_id={self.job_id}:")
            if results["inference_results"]["is_streaming"]:
                logger.info("Response:")
                for data in results["inference_results"]["response"]:
                    logger.info(data)
                self._is_consumed = True
            else:
                logger.info(json.dumps(results, ensure_ascii=False))

    def clean(self):
        pass

    @classmethod
    def _from_dict(cls, data) -> "SingleInferenceResult":
        return cls(
            job_id=data["job_id"],
            started_time=data["started_time"],
            inference_output_path="",
            nonstreaming_response=data["nonstreaming_response"],
            streaming_response=data["streaming_response"],
        )



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility to create custom RFT environments with boilerplate code.
"""

import os
import tarfile
from dataclasses import dataclass, field
from typing import Literal, Optional

import boto3

from amzn_nova_customization_sdk.util.logging import logger

from .custom_env_templates import (
    MULTI_TURN_TEMPLATE,
    PYPROJECT_TEMPLATE,
    README_TEMPLATE,
    SINGLE_TURN_TEMPLATE,
)


@dataclass
class CustomEnvironment:
    """Custom environment configuration"""

    env_id: str
    local_path: Optional[str] = None
    s3_uri: Optional[str] = None
    output_dir: str = "~/custom_envs"
    env_type: Literal["single_turn", "multi_turn"] = "single_turn"

    def create(self, overwrite: bool = False) -> "CustomEnvironment":
        """
        Create custom environment files with boilerplate code.

        Args:
            overwrite: Whether to overwrite existing directory

        Returns:
            self for chaining
        """
        if self.env_type not in ["single_turn", "multi_turn"]:
            raise ValueError(
                f"env_type must be 'single_turn' or 'multi_turn', got '{self.env_type}'"
            )

        env_name_safe = self.env_id.replace("_", "-").lower()
        module_name = env_name_safe.replace("-", "_")
        env_file = f"{module_name}.py"

        output_dir = os.path.expanduser(self.output_dir)
        env_dir = os.path.join(output_dir, env_name_safe)
        module_dir = os.path.join(env_dir, module_name)

        if os.path.exists(env_dir) and not overwrite:
            raise FileExistsError(
                f"Environment directory already exists: {env_dir}\nUse overwrite=True to replace it"
            )

        os.makedirs(module_dir, exist_ok=True)
        logger.info(f"Creating custom environment at: {env_dir}")

        # Write __init__.py to make it a package
        with open(os.path.join(module_dir, "__init__.py"), "w") as f:
            f.write(
                f"from .{module_name} import load_environment\n\n__all__ = ['load_environment']\n"
            )
        logger.info(f"Created __init__.py")

        # Write environment file in the package directory
        template = (
            SINGLE_TURN_TEMPLATE
            if self.env_type == "single_turn"
            else MULTI_TURN_TEMPLATE
        )
        with open(os.path.join(module_dir, env_file), "w") as f:
            f.write(template)
        logger.info(f"Created {module_name}/{env_file}")
        logger.warning(
            f"IMPORTANT: Please customize the generated {env_file} file with your own environment logic:"
        )
        logger.warning("   - Replace dataset loading with your own data source")
        logger.warning("   - Implement your custom reward function")
        logger.warning("   - Configure parser and rubric for your specific task")
        logger.warning(f"   See TODO comments in {os.path.join(module_dir, env_file)}")

        # Write pyproject.toml
        with open(os.path.join(env_dir, "pyproject.toml"), "w") as f:
            f.write(
                PYPROJECT_TEMPLATE.format(
                    env_name=env_name_safe, module_name=module_name
                )
            )
        logger.info("Created pyproject.toml")

        # Write README
        with open(os.path.join(env_dir, "README.md"), "w") as f:
            f.write(README_TEMPLATE.format(env_name=env_name_safe, env_path=env_dir))
        logger.info("Created README.md")

        # Set local_path
        self.local_path = env_dir

        logger.info(f"\nCustom environment '{env_name_safe}' created successfully!")

        return self

    def load(self) -> "CustomEnvironment":
        """
        Load existing custom environment from local_path.

        Returns:
            self for chaining

        Raises:
            ValueError: If local_path not set or doesn't exist
        """
        if not self.local_path:
            raise ValueError("local_path required to load environment")

        # Construct full path to environment directory
        base_path = os.path.expanduser(self.local_path)
        env_name_safe = self.env_id.replace("_", "-").lower()
        env_path = os.path.join(base_path, env_name_safe)

        if not os.path.exists(env_path):
            raise ValueError(
                f"Environment does not exist: {env_path}\n"
                f"Create it first with CustomEnvironment(...).create()"
            )

        if not os.path.isdir(env_path):
            raise ValueError(f"Environment path is not a directory: {env_path}")

        # Update local_path to point to the actual environment directory
        self.local_path = env_path

        logger.info(f"Loaded custom environment from: {env_path}")
        return self

    def validate(self) -> bool:
        """Validate custom environment has required structure"""
        if not self.local_path:
            raise ValueError("local_path required for validation")

        env_path = os.path.expanduser(self.local_path)

        if not os.path.exists(env_path):
            raise ValueError(f"Environment path does not exist: {env_path}")

        if not os.path.isdir(env_path):
            raise ValueError(f"Environment path is not a directory: {env_path}")

        # Check for pyproject.toml
        if not os.path.exists(os.path.join(env_path, "pyproject.toml")):
            raise ValueError(f"Missing pyproject.toml in {env_path}")

        # Check for Python module directory with load_environment
        # Look for subdirectories that could be the module
        subdirs = [
            d
            for d in os.listdir(env_path)
            if os.path.isdir(os.path.join(env_path, d))
            and not d.startswith(".")
            and not d.startswith("__")
        ]

        found_load_env = False
        for subdir in subdirs:
            subdir_path = os.path.join(env_path, subdir)
            py_files = [
                f
                for f in os.listdir(subdir_path)
                if f.endswith(".py") and not f.startswith("__")
            ]

            for py_file in py_files:
                with open(os.path.join(subdir_path, py_file), "r") as f:
                    if "def load_environment" in f.read():
                        found_load_env = True
                        break
            if found_load_env:
                break

        if not found_load_env:
            raise ValueError(f"No load_environment() function found in {env_path}")

        logger.info(f"Environment validation passed: {env_path}")
        return True

    def package_and_upload(
        self,
        s3_bucket: Optional[str] = None,
        s3_prefix: str = "custom-envs",
        region: str = "us-east-1",
    ) -> str:
        """
        Package custom environment and upload to S3 for ECS/EC2 deployment.

        Args:
            s3_bucket: S3 bucket name (optional, defaults to SageMaker default bucket)
            s3_prefix: S3 prefix/folder (default: "custom-envs")
            region: AWS region

        Returns:
            str: S3 URI of uploaded package
        """
        if not self.local_path:
            raise ValueError("local_path required for packaging")

        env_path = os.path.expanduser(self.local_path)
        self.validate()

        # Use SageMaker default bucket if not specified
        if not s3_bucket:
            import sagemaker

            session = sagemaker.Session(boto_session=boto3.Session(region_name=region))
            bucket_name = session.default_bucket()
            logger.info(f"Using SageMaker default bucket: {bucket_name}")
        else:
            if s3_bucket.startswith("s3://"):
                raise ValueError(
                    "s3_bucket should be bucket name only, not s3:// URI. Use s3_prefix for path."
                )
            bucket_name = s3_bucket

        tar_name = f"{self.env_id}.tar.gz"
        tar_path = os.path.join("/tmp", tar_name)

        logger.info(f"Packaging {env_path} to {tar_path}")
        with tarfile.open(tar_path, "w:gz") as tar:
            tar.add(env_path, arcname=self.env_id)

        s3_key = f"{s3_prefix.rstrip('/')}/{tar_name}" if s3_prefix else tar_name
        s3_uri = f"s3://{bucket_name}/{s3_key}"

        logger.info(f"Uploading to {s3_uri}")
        s3_client = boto3.client("s3", region_name=region)
        s3_client.upload_file(tar_path, bucket_name, s3_key)

        os.remove(tar_path)
        self.s3_uri = s3_uri

        logger.info(f"Custom environment uploaded: {s3_uri}")
        return s3_uri



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
SFT_NOVA_ONE_CONVERSE_2024 = {
    "type": "object",
    "properties": {
        "schemaVersion": {"type": "string", "const": "bedrock-conversation-2024"},
        "system": {
            "type": "array",
            "maxItems": 1,
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {"text": {"type": "string"}},
                "required": ["text"],
                "additionalProperties": False,
            },
        },
        "messages": {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {
                    "role": {"type": "string", "enum": ["user", "assistant"]},
                    "content": {
                        "type": "array",
                        "minItems": 1,
                        "items": {
                            "type": "object",
                            "oneOf": [
                                # Text content
                                {
                                    "properties": {"text": {"type": "string"}},
                                    "required": ["text"],
                                    "additionalProperties": False,
                                },
                                # Image content
                                {
                                    "properties": {
                                        "image": {
                                            "type": "object",
                                            "properties": {
                                                "format": {
                                                    "type": "string",
                                                    "enum": ["png", "jpg", "jpeg"],
                                                },
                                                "source": {
                                                    "type": "object",
                                                    "properties": {
                                                        "s3Location": {
                                                            "type": "object",
                                                            "properties": {
                                                                "uri": {
                                                                    "type": "string"
                                                                },
                                                                "bucketOwner": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                            "required": [
                                                                "uri",
                                                                "bucketOwner",
                                                            ],
                                                            "additionalProperties": False,
                                                        }
                                                    },
                                                    "required": ["s3Location"],
                                                    "additionalProperties": False,
                                                },
                                            },
                                            "required": ["format", "source"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["image"],
                                    "additionalProperties": False,
                                },
                                # Video content
                                {
                                    "properties": {
                                        "video": {
                                            "type": "object",
                                            "properties": {
                                                "format": {
                                                    "type": "string",
                                                    "enum": ["mp4", "mov", "avi"],
                                                },
                                                "source": {
                                                    "type": "object",
                                                    "properties": {
                                                        "s3Location": {
                                                            "type": "object",
                                                            "properties": {
                                                                "uri": {
                                                                    "type": "string"
                                                                },
                                                                "bucketOwner": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                            "required": [
                                                                "uri",
                                                                "bucketOwner",
                                                            ],
                                                            "additionalProperties": False,
                                                        }
                                                    },
                                                    "required": ["s3Location"],
                                                    "additionalProperties": False,
                                                },
                                            },
                                            "required": ["format", "source"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["video"],
                                    "additionalProperties": False,
                                },
                            ],
                        },
                    },
                },
                "required": ["role", "content"],
                "additionalProperties": False,
            },
        },
    },
    "required": ["schemaVersion", "messages"],
    "additionalProperties": False,
}

SFT_NOVA_TWO_CONVERSE_2024 = {
    "type": "object",
    "properties": {
        "schemaVersion": {"type": "string", "const": "bedrock-conversation-2024"},
        "system": {
            "type": "array",
            "maxItems": 1,
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {"text": {"type": "string"}},
                "required": ["text"],
                "additionalProperties": False,
            },
        },
        "messages": {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {
                    "role": {"type": "string", "enum": ["user", "assistant"]},
                    "content": {
                        "type": "array",
                        "minItems": 1,
                        "items": {
                            "type": "object",
                            "oneOf": [
                                # Text content
                                {
                                    "properties": {"text": {"type": "string"}},
                                    "required": ["text"],
                                    "additionalProperties": False,
                                },
                                # Reasoning content
                                {
                                    "properties": {
                                        "reasoningContent": {
                                            "type": "object",
                                            "properties": {
                                                "reasoningText": {
                                                    "type": "object",
                                                    "properties": {
                                                        "text": {"type": "string"}
                                                    },
                                                    "required": ["text"],
                                                    "additionalProperties": False,
                                                }
                                            },
                                            "required": ["reasoningText"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["reasoningContent"],
                                    "additionalProperties": False,
                                },
                                {
                                    "properties": {
                                        "toolUse": {
                                            "type": "object",
                                            "properties": {
                                                "toolUseId": {"type": "string"},
                                                "name": {"type": "string"},
                                                "input": {"type": "object"},
                                            },
                                            "required": ["toolUseId", "name", "input"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["toolUse"],
                                    "additionalProperties": False,
                                },
                                {
                                    "properties": {
                                        "toolResult": {
                                            "type": "object",
                                            "properties": {
                                                "toolUseId": {"type": "string"},
                                                "content": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "object",
                                                        "properties": {
                                                            "text": {"type": "string"}
                                                        },
                                                        "required": ["text"],
                                                        "additionalProperties": False,
                                                    },
                                                },
                                            },
                                            "required": ["toolUseId", "content"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["toolResult"],
                                    "additionalProperties": False,
                                },
                                # Image content
                                {
                                    "properties": {
                                        "image": {
                                            "type": "object",
                                            "properties": {
                                                "format": {
                                                    "type": "string",
                                                    "enum": ["jpeg", "jpg", "png"],
                                                },
                                                "source": {
                                                    "type": "object",
                                                    "properties": {
                                                        "s3Location": {
                                                            "type": "object",
                                                            "properties": {
                                                                "uri": {
                                                                    "type": "string"
                                                                },
                                                                "bucketOwner": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                            "required": [
                                                                "uri",
                                                                "bucketOwner",
                                                            ],
                                                            "additionalProperties": False,
                                                        }
                                                    },
                                                    "required": ["s3Location"],
                                                    "additionalProperties": False,
                                                },
                                            },
                                            "required": ["format", "source"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["image"],
                                    "additionalProperties": False,
                                },
                                # Video content
                                {
                                    "properties": {
                                        "video": {
                                            "type": "object",
                                            "properties": {
                                                "format": {
                                                    "type": "string",
                                                    "enum": ["mp4", "mov", "avi"],
                                                },
                                                "source": {
                                                    "type": "object",
                                                    "properties": {
                                                        "s3Location": {
                                                            "type": "object",
                                                            "properties": {
                                                                "uri": {
                                                                    "type": "string"
                                                                },
                                                                "bucketOwner": {
                                                                    "type": "string"
                                                                },
                                                            },
                                                            "required": [
                                                                "uri",
                                                                "bucketOwner",
                                                            ],
                                                            "additionalProperties": False,
                                                        }
                                                    },
                                                    "required": ["s3Location"],
                                                    "additionalProperties": False,
                                                },
                                            },
                                            "required": ["format", "source"],
                                            "additionalProperties": False,
                                        }
                                    },
                                    "required": ["video"],
                                    "additionalProperties": False,
                                },
                            ],
                        },
                    },
                },
                "required": ["role", "content"],
                "additionalProperties": False,
            },
        },
        "toolConfig": {
            "type": "object",
            "properties": {
                "tools": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "toolSpec": {
                                "type": "object",
                                "properties": {
                                    "name": {"type": "string"},
                                    "description": {"type": "string"},
                                    "inputSchema": {
                                        "type": "object",
                                        "properties": {"json": {"type": "object"}},
                                        "required": ["json"],
                                        "additionalProperties": False,
                                    },
                                },
                                "required": ["name", "inputSchema"],
                                "additionalProperties": False,
                            }
                        },
                        "required": ["toolSpec"],
                        "additionalProperties": False,
                    },
                }
            },
            "required": ["tools"],
            "additionalProperties": False,
        },
    },
    "required": ["schemaVersion", "messages"],
    "additionalProperties": False,
}

OPENAI_FORMAT = {
    "type": "object",
    "properties": {
        "messages": {
            "type": "array",
            "minItems": 1,
            "items": {
                "type": "object",
                "properties": {
                    "role": {
                        "type": "string",
                        "enum": ["system", "user", "assistant", "tool"],
                    },
                    "content": {
                        "type": ["string", "null"]
                    },  # Allow null for assistant messages with only tool_calls
                    "tool_calls": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string"},
                                "type": {"type": "string"},
                                "function": {
                                    "type": "object",
                                    "properties": {
                                        "name": {"type": "string"},
                                        "arguments": {"type": "string"},
                                    },
                                    "required": ["name", "arguments"],
                                },
                            },
                            "required": ["id", "type", "function"],
                        },
                    },
                    "tool_call_id": {"type": "string"},
                    "reasoning": {"type": "string"},
                },
                "required": ["role"],
                "additionalProperties": False,
            },
        },
        "tools": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "type": {"type": "string"},
                    "function": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                            "description": {"type": "string"},
                            "parameters": {"type": "object"},
                        },
                        "required": ["name"],
                    },
                },
                "required": ["type", "function"],
            },
        },
    },
    "required": ["messages"],
    "additionalProperties": False,
}

# TODO: Verify the core requirements for RFT formatting. Specifically "role" for names and what's required.
RFT_OPENAI_FORMAT = {
    "type": "object",
    "properties": {
        "messages": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "role": {
                        "type": "string",
                        "enum": ["system", "user", "assistant", "developer", "tool"],
                    },
                    "content": {
                        "type": ["string", "null"]
                    },  # Allow null for assistant messages with only tool_calls
                    "tool_calls": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string"},
                                "type": {"type": "string"},
                                "function": {
                                    "type": "object",
                                    "properties": {
                                        "name": {"type": "string"},
                                        "arguments": {"type": "string"},
                                    },
                                    "required": ["name", "arguments"],
                                },
                            },
                            "required": ["id", "type", "function"],
                        },
                    },
                    "tool_call_id": {"type": "string"},
                    "reasoning": {"type": "string"},
                },
                "required": ["role"],
            },
            "minItems": 1,
        },
        "reference_answer": {
            "type": "object",
            "additionalProperties": True,  # Allows any structure inside reference_answer
        },
        "tools": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "type": {"type": "string"},
                    "function": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string"},
                            "description": {"type": "string"},
                            "parameters": {"type": "object"},
                        },
                        "required": ["name"],
                    },
                },
                "required": ["type", "function"],
            },
        },
    },
    "required": ["messages"],
    "additionalProperties": True,  # This allows any additional top-level fields
}

EVALUATION_FORMAT = {
    "type": "object",
    "properties": {
        "query": {"type": "string"},
        "response": {"type": "string"},
        "system": {"type": "string"},
        "images": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"data": {"type": "string"}},
                "required": ["data"],
                "additionalProperties": False,
            },
        },
        "metadata": {"type": "string"},
    },
    "required": ["query", "response"],
    "additionalProperties": False,
}

CPT_FORMAT = {
    "type": "object",
    "properties": {
        "text": {"type": "string"},
    },
    "required": ["text"],
    "additionalProperties": False,
}



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

logging.basicConfig(level=logging.INFO)

# Suppress informational and warning logs from dependencies
logging.getLogger("botocore.credentials").setLevel(logging.ERROR)
logging.getLogger("sagemaker").setLevel(logging.ERROR)
logging.getLogger("sagemaker.config").setLevel(logging.ERROR)
logging.getLogger("sagemaker.pytorch").setLevel(logging.ERROR)

logger = logging.getLogger("nova_customization_sdk")
logger.setLevel(logging.INFO)
logger.propagate = True



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RFT Multiturn Infrastructure Module - Modular platform-specific implementations."""

from .base_infra import (
    RFT_EXECUTION_ROLE_NAME,
    EnvType,
    StackOutputs,
    VFEnvId,
    create_rft_execution_role,
)
from .custom_environment import CustomEnvironment
from .rft_multiturn import RFTMultiturnInfrastructure, list_rft_stacks

__all__ = [
    "RFTMultiturnInfrastructure",
    "list_rft_stacks",
    "create_rft_execution_role",
    "RFT_EXECUTION_ROLE_NAME",
    "EnvType",
    "VFEnvId",
    "StackOutputs",
    "CustomEnvironment",
]



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Helper functions for Sagemaker management.
"""

import json
import time
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

import boto3
from botocore.client import BaseClient
from botocore.exceptions import ClientError

from amzn_nova_customization_sdk.manager.runtime_manager import (
    RuntimeManager,
    SMHPRuntimeManager,
    SMTJRuntimeManager,
)
from amzn_nova_customization_sdk.model.model_config import (
    REGION_TO_ESCROW_ACCOUNT_MAPPING,
    ModelArtifacts,
)
from amzn_nova_customization_sdk.model.model_enums import DeploymentMode, Model
from amzn_nova_customization_sdk.model.result.inference_result import (
    SingleInferenceResult,
)
from amzn_nova_customization_sdk.validation.endpoint_validator import (
    validate_s3_uri_prefix,
)

from .logging import logger

# Model Parameters
DEFAULT_CONTEXT_LENGTH = "12000"
DEFAULT_MAX_CONCURRENCY = "16"

SAGEMAKER_EXECUTION_ROLE_NAME = "SageMakerDeployModelExecutionRole"


def _get_sagemaker_inference_image(region: str) -> str:
    if region not in REGION_TO_ESCROW_ACCOUNT_MAPPING:
        raise ValueError(
            f"Unsupported region: {region}. Supported regions are: {list(REGION_TO_ESCROW_ACCOUNT_MAPPING.keys())}"
        )

    return f"{REGION_TO_ESCROW_ACCOUNT_MAPPING[region]}.dkr.ecr.{region}.amazonaws.com/nova-inference-repo:SM-Inference-latest"


def get_model_artifacts(
    job_name: str, infra: RuntimeManager, output_s3_path: str
) -> ModelArtifacts:
    """
    Retrieve model artifacts for a job

    Args:
        job_name: Name of the job
        infra: Infrastructure of the job
        output_s3_path: Output S3 path of the job (only necessary for HyperPod)

    Returns:
        ModelArtifacts: Model artifact S3 paths

    Raises:
        Exception: If unable to obtain job artifact information
    """
    sagemaker_client = boto3.client("sagemaker")

    if isinstance(infra, SMTJRuntimeManager):
        response = sagemaker_client.describe_training_job(TrainingJobName=job_name)

        return ModelArtifacts(
            checkpoint_s3_path=response["CheckpointConfig"]["S3Uri"],
            output_s3_path=response["OutputDataConfig"]["S3OutputPath"],
        )
    # TODO: Figure out a reliable way to determine the RIG of a given job
    elif isinstance(infra, SMHPRuntimeManager):
        response = sagemaker_client.describe_cluster(ClusterName=infra.cluster_name)
        rigs = response.get("RestrictedInstanceGroups", [])

        # If there's only one RIG in the cluster, we know that the job had to be submitted to that RIG
        checkpoint_s3_path = None
        if len(rigs) == 1:
            checkpoint_s3_path = (
                rigs[0].get("EnvironmentConfig", {}).get("S3OutputPath")
            )

        return ModelArtifacts(
            checkpoint_s3_path=checkpoint_s3_path,
            output_s3_path=output_s3_path,
        )
    else:
        raise ValueError(f"Unsupported platform")


def get_cluster_instance_info(
    cluster_name: str, region: Optional[str] = None
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Get instance types and counts from a HyperPod cluster.

    Args:
        cluster_name: Name of the HyperPod cluster
        region: AWS region (optional, uses default session region if not provided)

    Returns:
        Dict with 'normal_instance_groups' and 'restricted_instance_groups' keys

    Raises:
        Exception: If unable to describe the cluster
    """
    if region is None:
        sagemaker_client = boto3.client("sagemaker")
    else:
        sagemaker_client = boto3.client("sagemaker", region_name=region)

    try:
        response = sagemaker_client.describe_cluster(ClusterName=cluster_name)

        normal_instance_groups = []
        restricted_instance_groups = []

        # Process normal instance groups
        for group in response.get("InstanceGroups", []):
            group_info = {
                "instance_group_name": group["InstanceGroupName"],
                "instance_type": group["InstanceType"],
                "current_count": group["CurrentCount"],
                "target_count": group["TargetCount"],
                "status": group["Status"],
            }
            normal_instance_groups.append(group_info)

        # Process restricted instance groups
        for group in response.get("RestrictedInstanceGroups", []):
            group_info = {
                "instance_group_name": group["InstanceGroupName"],
                "instance_type": group["InstanceType"],
                "current_count": group["CurrentCount"],
                "target_count": group["TargetCount"],
                "status": group["Status"],
            }
            restricted_instance_groups.append(group_info)

        return {
            "normal_instance_groups": normal_instance_groups,
            "restricted_instance_groups": restricted_instance_groups,
        }

    except Exception as e:
        raise RuntimeError(
            f"Failed to get cluster instance info for {cluster_name}: {str(e)}"
        )


def _get_hub_content(
    hub_name: str,
    hub_content_name: str,
    hub_content_type: str,
    region: str,
) -> Dict[str, Any]:
    """
     Get hub content from SageMaker via the DescribeHubContent API

    Args:
        hub_name: Name of the SageMaker Hub
        hub_content_name: Name of the hub content
        hub_content_type: Type of hub content
        region: AWS region

    Returns:
        Dict containing hub content
    """
    sagemaker_client = boto3.client("sagemaker", region_name=region)

    try:
        response = sagemaker_client.describe_hub_content(
            HubName=hub_name,
            HubContentType=hub_content_type,
            HubContentName=hub_content_name,
        )

        # Parse HubContentDocument if it's a JSON string
        if "HubContentDocument" in response:
            hub_content_document = response["HubContentDocument"]
            if isinstance(hub_content_document, str):
                try:
                    response["HubContentDocument"] = json.loads(hub_content_document)
                except (json.JSONDecodeError, TypeError):
                    # If parsing fails, leave the string as is
                    pass

    except Exception as e:
        raise RuntimeError(
            f"Failed to get SageMaker hub content for '{hub_content_name}': {str(e)}"
        )

    return response


# TODO: Update environment variables when variables finalized
def setup_environment_variables(
    context_length: str = DEFAULT_CONTEXT_LENGTH,
    max_concurrency: str = DEFAULT_MAX_CONCURRENCY,
    temperature: Optional[str] = None,
    top_p: Optional[str] = None,
    top_k: Optional[str] = None,
    max_new_tokens: Optional[str] = None,
    logprobs: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Set up environment variables for model configuration.

    Args:
        context_length (str, optional): Context length. Defaults to DEFAULT_CONTEXT_LENGTH.
        max_concurrency (str, optional): Maximum number of concurrency. Defaults to DEFAULT_MAX_CONCURRENCY.
        temperature (str, optional): Sampling temperature for text generation. Defaults to None.
        top_p (str, optional): Nucleus sampling probability threshold. Defaults to None.
        top_k (str, optional): Top-k sampling parameter. Defaults to None.
        max_new_tokens (str, optional): Maximum number of new tokens to generate. Defaults to None.
        logprobs (str, optional): Number of log probabilities to return. Defaults to None.

    Returns:
        Dict[str, Any]: A dictionary of environment variables for model configuration.
    """
    environment = {"CONTEXT_LENGTH": context_length, "MAX_CONCURRENCY": max_concurrency}

    # Add optional parameters if provided
    if temperature is not None:
        environment["DEFAULT_TEMPERATURE"] = temperature
    if top_p is not None:
        environment["DEFAULT_TOP_P"] = top_p
    if top_k is not None:
        environment["DEFAULT_TOP_K"] = top_k
    if max_new_tokens is not None:
        environment["DEFAULT_MAX_NEW_TOKENS"] = max_new_tokens
    if logprobs is not None:
        environment["DEFAULT_LOGPROBS"] = logprobs

    return environment


def _monitor_endpoint_creation(sagemaker_client: BaseClient, endpoint_name: str) -> str:
    """
    Monitors the status of a custom endpoint creation in SageMaker.

    Args:
        sagemaker_client: The boto3 sagemaker client used in the script
        endpoint_name: The name of the model endpoint.

    Returns:
        str: Final status of the model ('INSERVICE' or raises exception)
    """
    start_time = datetime.now(timezone.utc)

    while True:
        try:
            response = sagemaker_client.describe_endpoint(EndpointName=endpoint_name)
            status = response["EndpointStatus"]

            elapsed_time = datetime.now(timezone.utc) - start_time

            logger.info(f"Status: {status} | Elapsed: {elapsed_time}")

            if status.upper() == "INSERVICE":
                logger.info(
                    f"\n\nSUCCESS! Endpoint creation is complete! '{endpoint_name}' is now INSERVICE!"
                )
                logger.info(f"Total time elapsed: {elapsed_time}")
                return status
            elif status.upper() in ["FAILED"]:
                error_msg = (
                    f"\n\nERROR! Endpoint '{endpoint_name}' status is: {status}\n"
                )
                logger.error(
                    f"{error_msg}\nPlease check the AWS console for more details.\n"
                )
                raise Exception(error_msg)
        except Exception as e:
            logger.error(f"Error checking status: {str(e)}\n")
            raise
        time.sleep(60)  # Sleep for a minute.


# TODO: This will be replaced by Hub content as source of truth
def _validate_sagemaker_instance_type_for_model_deployment(
    instance_type: str, model: Model
) -> None:
    """
    Validation method that checks the instance_type and if it is compatible with the desired model

    Args:
        instance_type: instance type
        model: Model (enum)

    Raises:
        ValueError: If validation fails

    """

    accepted_configs = {
        Model.NOVA_MICRO: [
            "ml.g5.12xlarge",
            "ml.g6.12xlarge",
            "ml.g5.48xlarge",
            "ml.g6.48xlarge",
            "ml.p5.48xlarge",
        ],
        Model.NOVA_LITE: [
            "ml.g5.12xlarge",
            "ml.g6.12xlarge",
            "ml.g5.48xlarge",
            "ml.g6.48xlarge",
            "ml.p5.48xlarge",
        ],
        Model.NOVA_LITE_2: ["ml.p5.48xlarge"],
        Model.NOVA_PRO: ["ml.g6.48xlarge", "ml.p5.48xlarge"],
    }

    if instance_type not in accepted_configs[model]:
        raise ValueError(
            f"{instance_type} is not in the supported instances list for {model}: {accepted_configs[model]}"
        )


def create_model_and_endpoint_config(
    region: str,
    model_name: str,
    model_s3_location: str,
    sagemaker_execution_role_arn: str,
    endpoint_config_name: str,
    endpoint_name: str,
    sagemaker_client: BaseClient,
    deployment_mode: DeploymentMode = DeploymentMode.FAIL_IF_EXISTS,
    instance_type: Optional[str] = "ml.g5.4xlarge",
    environment: Dict[str, Any] = {},
    initial_instance_count: Optional[int] = 1,
    network_isolation: bool = True,
) -> str:
    """
    Create a SageMaker model, endpoint configuration, and endpoint.
    If DeploymentMode is FAIL_IF_EXISTS, deployment will fail if model, endpoint configuration or endpoint already exist.
    DeploymentMode UPDATE_IF_EXISTS is not supported as model and endpoint configuration do not support updates.

    Args:
        region (str): AWS region
        model_name (str): Name of the SageMaker model.
        model_s3_location (str): S3 URI where the model artifacts are stored.
        sagemaker_execution_role_arn (str): IAM role ARN for SageMaker execution.
        endpoint_config_name (str): Name for the endpoint configuration.
        endpoint_name (str): Name for the SageMaker endpoint.
        instance_type (str): EC2 instance type for the endpoint.
        environment (Dict[str, Any]): Environment variables for the model.
        sagemaker_client (BaseClient): SageMaker client
        deployment_mode (DeploymentMode): How to handle existing model, endpoint configs and endpoints
        initial_instance_count (int, optional): Number of instances for the endpoint. Defaults to 1.
        network_isolation (bool, optional): Enable network isolation. Defaults to True.
    Returns:
        str: endpoint ARN

    Raises:
        Exception: If there's an error creating the model, endpoint config, or endpoint.
    """

    validate_s3_uri_prefix(s3_uri=model_s3_location)

    try:
        # Check for existing resources based on deployment mode
        existing_model = None
        existing_endpoint_config = None
        existing_endpoint = None

        try:
            existing_model = sagemaker_client.describe_model(ModelName=model_name)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ValidationException":
                pass
            else:
                raise

        try:
            existing_endpoint_config = sagemaker_client.describe_endpoint_config(
                EndpointConfigName=endpoint_config_name
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ValidationException":
                pass
            else:
                raise

        try:
            existing_endpoint = sagemaker_client.describe_endpoint(
                EndpointName=endpoint_name
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ValidationException":
                pass
            else:
                raise

        # Handle existing resources based on deployment mode
        if deployment_mode in [
            DeploymentMode.FAIL_IF_EXISTS,
            DeploymentMode.UPDATE_IF_EXISTS,
        ]:
            if existing_model:
                raise Exception(f"Model '{model_name}' already exists.")
            if existing_endpoint_config:
                raise Exception(
                    f"Endpoint configuration '{endpoint_config_name}' already exists."
                )
            if existing_endpoint:
                raise Exception(f"Endpoint '{endpoint_name}' already exists.")

        logger.info(f"Creating model: {model_name}...")
        model_response = sagemaker_client.create_model(
            ModelName=model_name,
            PrimaryContainer={
                "Image": _get_sagemaker_inference_image(region),
                "ModelDataSource": {
                    "S3DataSource": {
                        "S3Uri": model_s3_location,
                        "S3DataType": "S3Prefix",
                        "CompressionType": "None",
                    }
                },
                "Environment": environment,
            },
            ExecutionRoleArn=sagemaker_execution_role_arn,
            EnableNetworkIsolation=network_isolation,
        )
        logger.info(f"Model created successfully: {model_response['ModelArn']}")

        production_variant = {
            "VariantName": "primary",
            "ModelName": model_name,
            "InitialInstanceCount": initial_instance_count,
            "InstanceType": instance_type,
        }

        logger.info(f"Creating endpoint configuration: {endpoint_config_name}...")
        # Create endpoint configuration
        config_response = sagemaker_client.create_endpoint_config(
            EndpointConfigName=endpoint_config_name,
            ProductionVariants=[production_variant],
        )
        logger.info(
            f"Endpoint configuration created successfully: {config_response['EndpointConfigArn']}"
        )

        logger.info(f"Creating endpoint: {endpoint_name}...")
        endpoint_response = sagemaker_client.create_endpoint(
            EndpointName=endpoint_name, EndpointConfigName=endpoint_config_name
        )

        # Poll for endpoint status
        logger.info(
            "Waiting for endpoint creation to complete. This can take ~10 minutes..."
        )
        try:
            _monitor_endpoint_creation(sagemaker_client, endpoint_name)
        except Exception as e:
            raise Exception(f"Failed to create deployment {endpoint_name}: {e}")

        return endpoint_response["EndpointArn"]

    except Exception as e:
        logger.info(f"Error creating model and endpoint: {e}")
        raise


def invoke_sagemaker_inference(
    request_body: Dict[str, Any],
    endpoint_name: str,
    sagemaker_client: BaseClient,
) -> SingleInferenceResult:
    """
     Invoke Sagemaker inference and return result

    Args:
        request_body (Dict[str, Any]): The payload to send to the inference endpoint.
        endpoint_name (str): Name of the SageMaker inference endpoint.
        sagemaker_client (BaseClient): Sagemaker client

    Returns:
        - Generator[str, None, None] for streaming responses
        - str for non-streaming responses
    """
    current_time = datetime.now(timezone.utc)

    body = json.dumps(request_body)
    is_streaming = request_body.get("stream", False)

    try:
        logger.info(
            f"Invoking endpoint ({'streaming' if is_streaming else 'non-streaming'})..."
        )

        if is_streaming:
            response = sagemaker_client.invoke_endpoint_with_response_stream(
                EndpointName=endpoint_name, ContentType="application/json", Body=body
            )

            event_stream = response["Body"]

            def stream_generator():
                for event in event_stream:
                    if "PayloadPart" in event:
                        chunk = event["PayloadPart"]
                        if "Bytes" in chunk:
                            data = chunk["Bytes"].decode()
                            yield data

            return SingleInferenceResult(
                job_id=response["ResponseMetadata"]["RequestId"],
                inference_output_path="",
                started_time=current_time,
                streaming_response=stream_generator(),
                nonstreaming_response=None,
            )
        else:
            response = sagemaker_client.invoke_endpoint(
                EndpointName=endpoint_name,
                ContentType="application/json",
                Accept="application/json",
                Body=body,
            )

            body_content = json.loads(response["Body"].read().decode("utf-8"))

            return SingleInferenceResult(
                job_id=body_content["id"],
                inference_output_path="",
                started_time=datetime.fromtimestamp(body_content["created"]),
                streaming_response=None,
                nonstreaming_response=body_content["choices"],
            )

    except Exception as e:
        raise Exception(f"Error invoking endpoint {endpoint_name}: {str(e)}")



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import shutil
import tarfile
import tempfile
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional
from urllib.parse import urlparse

import boto3

from amzn_nova_customization_sdk.model.result.job_result import (
    BaseJobResult,
    JobStatus,
    JobStatusManager,
    SMHPStatusManager,
    SMTJStatusManager,
)
from amzn_nova_customization_sdk.recipe.recipe_config import EvaluationTask
from amzn_nova_customization_sdk.util.logging import logger


@dataclass
class EvaluationResult(BaseJobResult, ABC):
    eval_task: EvaluationTask
    eval_output_path: str

    def __init__(
        self,
        job_id: str,
        started_time: datetime,
        eval_task: EvaluationTask,
        eval_output_path: str,
        s3_client=None,
    ):
        self.eval_task = eval_task
        self.eval_output_path = eval_output_path
        self._cached_results_dir: Optional[str] = None
        self._s3_client = s3_client or boto3.client("s3")
        super().__init__(job_id, started_time)

    def _download_eval_results(self) -> str:
        # Check if results are already cached
        if self._cached_results_dir and Path(self._cached_results_dir).exists():
            return self._cached_results_dir
        try:
            logger.info(f"Downloading results from {self.eval_output_path}")
            parsed = urlparse(self.eval_output_path)
            bucket = parsed.netloc
            key = parsed.path.lstrip("/")

            # Create temp dir for caching results
            self._cached_results_dir = tempfile.mkdtemp(
                prefix=f"eval_results_{self.job_id}_"
            )

            # Check if it's a tar.gz file or S3 directory
            if self.eval_output_path.endswith(".tar.gz"):
                # Handle tar.gz file
                local_file = Path(self._cached_results_dir) / "output.tar.gz"
                self._s3_client.download_file(bucket, key, str(local_file))
                with tarfile.open(local_file, "r:gz") as tar:
                    tar.extractall(self._cached_results_dir, filter="data")
            else:
                # Handle S3 directory
                paginator = self._s3_client.get_paginator("list_objects_v2")
                for page in paginator.paginate(Bucket=bucket, Prefix=key):
                    for obj in page.get("Contents", []):
                        obj_key = obj["Key"]
                        relative_path = obj_key[len(key) :].lstrip("/")
                        if relative_path:  # Skip empty paths
                            local_path = Path(self._cached_results_dir) / relative_path
                            local_path.parent.mkdir(parents=True, exist_ok=True)
                            self._s3_client.download_file(
                                bucket, obj_key, str(local_path)
                            )

            logger.info(
                f"Successfully downloaded eval results to {self._cached_results_dir}"
            )
            return self._cached_results_dir
        except Exception as e:
            print(f"Error retrieving evaluation results: {e}")
            print(f"Results available at: {self.eval_output_path}")
            raise

    def get(self) -> Dict:
        job_status, raw_status = self.get_job_status()
        if job_status == JobStatus.COMPLETED:
            print(f"Job '{self.job_id}' completed successfully.")
            results_dir = self._download_eval_results()
            results_files = list(Path(results_dir).rglob("results_*.json"))
            if results_files:
                with open(results_files[0], "r") as f:
                    return json.load(f)
            else:
                logger.info(
                    f"No evaluation results json file found for job {self.job_id}"
                )
                print(f"No evaluation results json file found for job {self.job_id}")
                return {}
        elif job_status == JobStatus.IN_PROGRESS:
            print(
                f"Job '{self.job_id}' still running in progress. Please wait until the job is completed."
            )
            return {}
        else:
            print(
                f"Cannot show eval result. Job '{self.job_id}' in {raw_status} status."
            )
            return {}

    def show(self):
        results = self.get()
        if results:
            print(
                f"\nEvaluation Results for job_id={self.job_id}, eval_task={self.eval_task.value}:"
            )
            print(json.dumps(results, indent=2))

    def upload_tensorboard_results(self, tensorboard_s3_path: Optional[str] = None):
        """
        Upload eval tensorboard result to s3
        :param tensorboard_s3_path: Optional, the s3 path you want the tensorboard result upload to. Will use
        eval_output_path if not provide.
        :return: None
        """
        # Ensure results are downloaded first
        results_dir = self._download_eval_results()

        # Search for tensorboard_results directory in subdirectories
        tensorboard_dirs = list(Path(results_dir).rglob("tensorboard_results"))
        if not tensorboard_dirs:
            logger.warning(
                f"No tensorboard_results directory found in {results_dir} or its subdirectories"
            )
            print(
                f"No tensorboard_results directory found in {results_dir} or its subdirectories"
            )
            return

        tensorboard_dir = tensorboard_dirs[0]  # Use the first found directory

        # Determine target S3 path
        if tensorboard_s3_path is None:
            # Replace output.tar.gz with tensorboard_results/
            eval_s3_path = self.eval_output_path
            tensorboard_s3_path = eval_s3_path.replace(
                "/output.tar.gz", "/tensorboard_results/"
            )

        # Parse S3 path
        parsed = urlparse(tensorboard_s3_path)
        bucket = parsed.netloc
        key_prefix = parsed.path.lstrip("/")

        # Upload all files in tensorboard_results directory
        for file_path in tensorboard_dir.rglob("*"):
            if file_path.is_file():
                relative_path = file_path.relative_to(tensorboard_dir)
                s3_key = f"{key_prefix}{relative_path}"

                try:
                    self._s3_client.upload_file(str(file_path), bucket, s3_key)
                    logger.debug(f"Uploaded {file_path} to s3://{bucket}/{s3_key}")
                except Exception as e:
                    logger.error(f"Failed to upload {file_path}: {e}")
                    raise

        logger.info(
            f"Successfully uploaded tensorboard results to {tensorboard_s3_path}"
        )
        print(f"Successfully uploaded tensorboard results to {tensorboard_s3_path}")

    def clean(self):
        """Clean up local results cache"""
        if (
            hasattr(self, "_cached_results_dir")
            and self._cached_results_dir
            and Path(self._cached_results_dir).exists()
        ):
            try:
                shutil.rmtree(self._cached_results_dir)
                logger.debug(
                    f"Successfully cleaned temp file in {self._cached_results_dir}"
                )
                self._cached_results_dir = None
            except Exception as e:
                logger.error(
                    f"Failed to remove cached results directory: {self._cached_results_dir} due to {str(e)}"
                )

    def __del__(self):
        # Clean the cache during GC
        self.clean()


@dataclass
class SMTJEvaluationResult(EvaluationResult):
    def __init__(
        self,
        job_id: str,
        started_time: datetime,
        eval_task: EvaluationTask,
        eval_output_path: str,
        sagemaker_client=None,
        s3_client=None,
    ):
        self._sagemaker_client = sagemaker_client or boto3.client("sagemaker")
        super().__init__(job_id, started_time, eval_task, eval_output_path, s3_client)

    def _create_status_manager(self) -> JobStatusManager:
        return SMTJStatusManager(self._sagemaker_client)

    def _to_dict(self):
        return {
            "job_id": self.job_id,
            "started_time": self.started_time.isoformat(),
            "eval_task": self.eval_task.value,
            "eval_output_path": self.eval_output_path,
        }

    @classmethod
    def _from_dict(cls, data) -> "SMTJEvaluationResult":
        return cls(
            job_id=data["job_id"],
            started_time=datetime.fromisoformat(data["started_time"]),
            eval_task=EvaluationTask(data["eval_task"]),
            eval_output_path=data["eval_output_path"],
        )


@dataclass
class SMHPEvaluationResult(EvaluationResult):
    cluster_name: str
    namespace: str

    def __init__(
        self,
        job_id: str,
        started_time: datetime,
        eval_task: EvaluationTask,
        eval_output_path: str,
        cluster_name: str,
        namespace: str = "kubeflow",
    ):
        self.cluster_name = cluster_name
        self.namespace = namespace
        super().__init__(job_id, started_time, eval_task, eval_output_path)

    def _create_status_manager(self) -> JobStatusManager:
        return SMHPStatusManager(self.cluster_name, self.namespace)

    def _to_dict(self):
        return {
            "job_id": self.job_id,
            "started_time": self.started_time.isoformat(),
            "eval_task": self.eval_task.value,
            "eval_output_path": self.eval_output_path,
            "cluster_name": self.cluster_name,
            "namespace": self.namespace,
        }

    @classmethod
    def _from_dict(cls, data) -> "SMHPEvaluationResult":
        return cls(
            job_id=data["job_id"],
            started_time=datetime.fromisoformat(data["started_time"]),
            eval_task=EvaluationTask(data["eval_task"]),
            eval_output_path=data["eval_output_path"],
            cluster_name=data["cluster_name"],
            namespace=data["namespace"],
        )



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data models for Nova Customization SDK.

This module contains dataclass definitions and constants used across the SDK.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional, TypedDict

from amzn_nova_customization_sdk.model.model_enums import DeployPlatform

REGION_TO_ESCROW_ACCOUNT_MAPPING = {
    "us-east-1": "708977205387",
    "us-west-2": "176779409107",
    "eu-west-2": "470633809225",
}


class ModelConfigDict(TypedDict):
    type: str
    path: str


@dataclass
class ModelArtifacts:
    checkpoint_s3_path: Optional[str]
    output_s3_path: str


@dataclass
class EndpointInfo:
    platform: DeployPlatform
    endpoint_name: str
    uri: str
    model_artifact_path: str


@dataclass
class DeploymentResult:
    endpoint: EndpointInfo
    created_at: datetime

    @property
    def status(self):
        from amzn_nova_customization_sdk.util.bedrock import check_deployment_status

        return check_deployment_status(self.endpoint.uri, self.endpoint.platform)



# Copyright 2025 Amazon Inc

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Data mixing configuration for Nova models.

This module provides the DataMixing dataclass for configuring data mixing
between customer data and Nova curated data for SFT training.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from amzn_nova_customization_sdk.util.logging import logger
from amzn_nova_customization_sdk.validation.validator import Validator


@dataclass
class DataMixing:
    """
    Data mixing configuration for Nova models.

    This class manages datamixing configuration for SFT training, including
    validation and format conversion. It provides methods to get, set, and
    validate data mixing configurations.

    Attributes:
        config: Dictionary containing the datamixing configuration
                Keys and values are dynamically determined from recipe templates
        _default_nova_fields: Set of known nova field names for normalization
        _dataset_catalog: Stores the dataset catalog value (read-only from template)
    """

    # Constants for field name patterns
    NOVA_PREFIX = "nova_"
    PERCENT_SUFFIX = "_percent"
    CUSTOMER_DATA_FIELD = "customer_data_percent"
    DATASET_CATALOG_FIELD = "dataset_catalog"

    config: Dict[str, Any] = field(default_factory=dict)
    _default_nova_fields: set = field(
        default_factory=lambda: set(), init=False, repr=False
    )
    _dataset_catalog: Optional[Any] = field(default=None, init=False, repr=False)

    def get_config(self) -> Dict[str, Any]:
        """
        Get the current data mixing configuration.

        Returns:
            A copy of the current data mixing configuration dictionary including dataset_catalog
        """
        result = self.config.copy()
        if self._dataset_catalog is not None:
            result[self.DATASET_CATALOG_FIELD] = self._dataset_catalog
        return result

    def set_config(self, config: Dict[str, Any], normalize: bool = True) -> None:
        """
        Set the data mixing configuration.

        Args:
            config: Dictionary containing the data mixing configuration.
                   Keys should include nova_*_percent fields and customer_data_percent.
                   Any nova_*_percent fields not specified will be set to 0 if normalize=True.
            normalize: If True, unspecified nova fields will be set to 0. Default is True.

        Raises:
            ValueError: If configuration is invalid or contains unknown nova fields
        """
        if self.DATASET_CATALOG_FIELD in config:
            logger.warning(
                f"{self.DATASET_CATALOG_FIELD} cannot be set in data mixing configuration. "
                f"Ignoring value: {config[self.DATASET_CATALOG_FIELD]}"
            )

        # Create new config with normalization if needed
        new_config = {}

        if normalize:
            # Start with all known fields set to 0
            if (
                self._default_nova_fields
                or self.CUSTOMER_DATA_FIELD in self._default_nova_fields
            ):
                # Include nova fields
                for field in self._default_nova_fields:
                    new_config[field] = 0
                # Include customer_data_percent if it's in defaults
                if self.CUSTOMER_DATA_FIELD in self._default_nova_fields:
                    new_config[self.CUSTOMER_DATA_FIELD] = 0

        # Update with provided config
        new_config.update(config)

        # Validate the configuration
        Validator.validate_data_mixing_config(
            new_config,
            self.NOVA_PREFIX,
            self.PERCENT_SUFFIX,
            self.CUSTOMER_DATA_FIELD,
            self.DATASET_CATALOG_FIELD,
            self._default_nova_fields,
        )

        # Store the new config
        self.config = new_config

        logger.info(
            f"Setting other data mixing fields to 0. Data mixing configuration set to: {self.config}"
        )

    def _load_defaults_from_template(self, overrides_template: Dict[str, Any]) -> None:
        """
        Load default configuration from an overrides template.

        Args:
            overrides_template: Template dictionary containing default values
        """
        default_config = {}
        for key, value in overrides_template.items():
            if key.startswith(self.NOVA_PREFIX) and key.endswith(self.PERCENT_SUFFIX):
                self._default_nova_fields.add(key)
                if isinstance(value, dict) and "default" in value:
                    default_config[key] = value["default"]
            elif key == self.CUSTOMER_DATA_FIELD:
                # Add customer_data_percent to default fields
                self._default_nova_fields.add(key)
                if isinstance(value, dict) and "default" in value:
                    default_config[key] = value["default"]
            elif key == self.DATASET_CATALOG_FIELD:
                # Store dataset_catalog separately as it's read-only
                if isinstance(value, dict) and "default" in value:
                    self._dataset_catalog = value["default"]

        self.config.update(default_config)

    def _is_data_mixing_field(self, key: str) -> bool:
        return (
            (
                key.startswith(DataMixing.NOVA_PREFIX)
                and key.endswith(DataMixing.PERCENT_SUFFIX)
            )
            or key == DataMixing.CUSTOMER_DATA_FIELD
            or DataMixing.NOVA_PREFIX + key + DataMixing.PERCENT_SUFFIX
            in self.get_config()
            or key == DataMixing.DATASET_CATALOG_FIELD
        )
